//******************************************************************************
//*                                 uBeeDisk                                   *
//*                                                                            *
//*     A tool for converting disks/images from one to another with auto       *
//*                   detection options for Microbee disks.                    *
//*                                                                            *
//*                              Applix module                                 *
//*                                                                            *
//*                   Copyright (C) 2008-2014 Stewart Kay                      *
//******************************************************************************
//
// Probe for known Applix disk formats.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.1.0 - 8 October 2011, S.J.Kay
// - Initial file created from formats.c

//==============================================================================
/*
 *  uBeeDisk - A tool for converting disks/images from one to another with
 *  auto detection options for Microbee disks.
 *  Copyright (C) 2008-2014 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdint.h>

#include <libdsk.h>

#include "format.h"
#include "applix.h"
#include "ubeedisk.h"
#include "functions.h"

//==============================================================================
// structures and variables
//
// applix_disk_format[]
// ------------------
// The order in applix_disk_format[] must match the fmt_* enumeration in
// applix.h
//
// The skew_val and skew_ofs value should be set to -1 if the default values
// should be used.
//
// The GAP values are based on the information for 5x1024 found here:
// http://info-coach.fr/atari/software/FD-Soft.php
//==============================================================================
disk_format_t applix_disk_format[] =
{
 {"applix80", "Applix 1616/OS DS DD 80T 5x1024 b/s 3.5\"",
  SIDES_ALT, 80, 2,  5, 1, 1024, RATE_SD, 40, 120, 0, 1, 1, 0, -1,  1,  1, 0, 
  FLAGS_APPLIX},

 {"applix81", "Applix 1616/OS DS DD 81T 5x1024 b/s 3.5\"",
  SIDES_ALT, 81, 2,  5, 1, 1024, RATE_SD, 40, 120, 0, 1, 1, 0, -1,  1,  1, 0, 
  FLAGS_APPLIX},

 {"applix82", "Applix 1616/OS DS DD 82T 5x1024 b/s 3.5\"",
  SIDES_ALT, 82, 2,  5, 1, 1024, RATE_SD, 40, 120, 0, 1, 1, 0, -1,  1,  1, 0, 
  FLAGS_APPLIX},

 {"applix83", "Applix 1616/OS DS DD 83T 5x1024 b/s 3.5\"",
  SIDES_ALT, 83, 2,  5, 1, 1024, RATE_SD, 40, 120, 0, 1, 1, 0, -1,  1,  1, 0, 
  FLAGS_APPLIX},

 {"", "", 
          0,  0, 0,  0, 0,    0,       0,  0,  0, 0, 0, 0, 0,  0,  0, -1,-1, 0}
};

extern char *no_dsk_ptrackids[];

extern DSK_PDRIVER idrive;
extern DSK_GEOMETRY dg;
extern disk_t disk;
extern xdg_t xdg;

//==============================================================================
// Detect Applix disk/image format.
//
// The first 2 bytes of the disk image records the number of 1024 byte
// blocks on the disk.  This is used to determine how many tracks are on the
// disk so the correct format can be set.
//
// Sector skewing Side 0:
// 42531
//
// Sector skewing Side 1:
// 14253
//
//   pass: void
// return: int				0 if successful else -1
//==============================================================================
int applix_detect_format (void)
{
 char buffer[1024];
 int blocks;
 int f;
 
 // values common to all Applix formats?
 format_set_geometry(&applix_disk_format[fmt_applix80]);

 // read first sector from T0, H0
 if (read_physical_sector(0, 0, 0, 0, 1, buffer) != DSK_ERR_OK)
    {
     printf("\nApplix disk detection not able to read sector.\n");
     strcpy(xdg.dg_format_name, "");
     return -1;
    }

 // number of blocks on disk (big endian format)
 blocks = buffer[0] << 8 | buffer[1];
 
 switch (blocks)
    {
     case 800 :
        f = fmt_applix80;
        break;
     case 810 :
        f = fmt_applix81;
        break;
     case 820 :
        f = fmt_applix82;
        break;
     case 830 :
        f = fmt_applix83;
        break;     
      default :
         strcpy(xdg.dg_format_name, "");
         printf("\nApplix block size (%d) not recognised.\n", blocks);
         return -1;
    }

 // set the format
 format_set_geometry(&applix_disk_format[f]);

 // report the format that has been set
 if (disk.verbose)
    printf("%s: %s\n", applix_disk_format[f].name,
           applix_disk_format[f].desc);

 return 0;
}
