//******************************************************************************
//*                                 uBeeDisk                                   *
//*                                                                            *
//*     A tool for converting disks/images from one to another with auto       *
//*                   detection options for Microbee disks.                    *
//*                                                                            *
//*                                fm module                                   *
//*                                                                            *
//*                   Copyright (C) 2008-2014 Stewart Kay                      *
//******************************************************************************
//
// Probe for known FM disk disk formats (8" disks).
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 30 August 2012, S.J.Kay
// - Commented out assigning unused variable dsk_err_h1 from fm_detect_format().
//
// v2.1.0 - 8 October 2011, S.J.Kay
// - Initial file created from formats.c

//==============================================================================
/*
 *  uBeeDisk - A tool for converting disks/images from one to another with
 *  auto detection options for Microbee disks.
 *  Copyright (C) 2008-2014 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdint.h>

#include <libdsk.h>

#include "format.h"
#include "fm.h"
#include "ubeedisk.h"
#include "functions.h"

//==============================================================================
// structures and variables
//
// fm_disk_format[]
// ----------------
// The order in fm_disk_format[] must match the fmt_* enumeration in
// format.h
//
// The skew_val and skew_ofs value should be set to -1 if the default values
// should be used.
//==============================================================================
disk_format_t fm_disk_format[] =
{
 {"ibm3740", "8in single sided, single density, 77 tracks, 26x128 b/s",
  SIDES_ALT, 77, 1, 26, 1,  128, RATE_HD,  0,  0, 1, 1, 1, 0, -1,  1,  1, 0, 0},
 {"gbss77",  "8in single sided, single density, 77 tracks, 9x512 b/s",
  SIDES_ALT, 77, 1,  9, 1,  512, RATE_HD,  0,  0, 1, 1, 1, 0, -1,  1, -1,-1, 0},
 {"3030", "30 Sector - 8in 77T DS DD 30x128 b/s",
  SIDES_ALT, 77, 1, 30, 1,  128, RATE_HD,  0,  0, 1, 1, 1, 0, -1,  1, -1,-1, 0},
 {"60", "60 Sector - 8in 77T DS DD 30x128 b/s",
  SIDES_ALT, 77, 2, 30, 1,  128, RATE_HD,  0,  0, 1, 1, 1, 0, -1,  1, -1,-1,
  FLAGS_60},
 {"", "", 
          0,  0, 0,  0, 0,    0,       0,  0,  0, 0, 0, 0, 0,  0,  0, -1,-1, 0}
};

extern char *no_dsk_ptrackids[];

extern DSK_PDRIVER idrive;
extern DSK_GEOMETRY dg;
extern disk_t disk;
extern xdg_t xdg;

//==============================================================================
// Detect Other disk/image format.
//
//   pass: void
// return: int				0 if successful else -1
//==============================================================================
int fm_detect_format (void)
{
 dsk_err_t dsk_err_h0 = DSK_ERR_OK;
// dsk_err_t dsk_err_h1 = DSK_ERR_OK;

 dsk_psect_t count_h0;
 dsk_psect_t count_h1;

 int size_h0;
 int size_h1;
 
 int t0h0_sec_f = 0;
 int t0h0_sec_l = 0;
 int t0h1_sec_f = 0;
 int t0h1_sec_l = 0;
 
 int sizeok_h0;
 int sizeok_h1;

 int f = -1; 
 int drive_type = 0;

 // establish the drive type detection required
 if ((disk.idrive_type == 'd') || (disk.idrive_type == 'h'))
    drive_type = 5;
 else
    drive_type = 3;

 // set some values depending on the drive type
 if (drive_type == 5)
    format_set_geometry(&fm_disk_format[fmt_ibm3740]);
 else   
    {
     printf("\nNo FM formats for 3.5\" drives implemented.\n");
     return -1;
    } 

 // read cylinder 0, side 0 of the disk
 dsk_err_h0 = read_sector_ids(0, 0, 128, &sizeok_h0, &size_h0,
 &count_h0, &t0h0_sec_f, &t0h0_sec_l);

 // read cylinder 0, side 1 of the disk
// dsk_err_h1 = 
 read_sector_ids(0, 1, 128, &sizeok_h1, &size_h1,
 &count_h1, &t0h1_sec_f, &t0h1_sec_l);

 // if nothing on T0 H0 then bad, non formatted, no disk, wrong data rate ?
 if (dsk_err_h0 != DSK_ERR_OK)
    {
     printf("No disk, non formatted, bad disk, or wrong data rate ???\n");
     return -1;
    }

 // check values
 for (;;)
    {
     if (size_h0 == 128 && count_h0 == 26)
        {
         f = fmt_ibm3740;
         break;
        } 
     if (size_h0 == 512 && count_h0 == 9)
        {
         f = fmt_gbss77;
         break;
        } 
     if (size_h0 == 128 && count_h0 == 30 && size_h1 == 128 && t0h1_sec_f < 31)
        {
         f = fmt_3030;
         break;
        } 
     if (size_h0 == 128 && count_h0 == 30 && size_h1 == 128 && t0h1_sec_f > 30)
        {
         f = fmt_60;
         break;
        } 

     // no format detected
     break;
    }
     
 // set the format
 if (f != -1)
    {
     format_set_geometry(&fm_disk_format[f]);

     // report the format that has been set
     if (disk.verbose)
        printf("%s: %s\n", fm_disk_format[f].name, fm_disk_format[f].desc);
     return 0;
    }

 printf("\nFormat not recognised.\n");
 printf(
 "Cylinder 0, Head 0: Sectors/track: %d  Size: %d  First: %d  Last: %d\n",
 count_h0, size_h0, t0h0_sec_f, t0h0_sec_l);
 printf(
 "Cylinder 0, Head 1: Sectors/track: %d  Size: %d  First: %d  Last: %d\n",
 count_h1, size_h1, t0h1_sec_f, t0h1_sec_l);

 strcpy(xdg.dg_format_name, "");
 return -1;
}
