//******************************************************************************
//*                                 uBeeDisk                                   *
//*                                                                            *
//*     A tool for converting disks/images from one to another with auto       *
//*                   detection options for Microbee disks.                    *
//*                                                                            *
//*                            functions module                                *
//*                                                                            *
//*                   Copyright (C) 2008-2014 Stewart Kay                      *
//******************************************************************************
//
// Various functions that don't belong anywhere else.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.2.0 - 14 May 2012, S.J.Kay
// - Added xdrive parameter to get_drive_type() and code to enquire from
//   remote devices what the drive type is.
// - Added function_init() and function_deinit(). These currently
//   pre-configure and reinstate Unix stdin values.
// - Change Unix get_key() to use the more modern termios methods and
//   different configurations to prevent unwanted key echo.
// - Added a fgetsx() wrapper for fgets() and printfx() wrapper for printf()
//   which temporarily enables stdout echo under Unices.
//
// v2.0.0 - 25 September 2011, S.J.Kay
// - Added get_key() function and support code for Unix POSIX and windows
//   systems to test for a key using non-blocking methods.
// - Added time_get_ms() and sleep_ms() functions.
// - Added get_integer_value() and get_next_parameter() functions.
//
// v1.4.0 - 24 February 2010, S.J.Kay
// - Added toupper_string() and tolower_string() functions.
//
// v1.0.0 - 22 December 2008, S.J.Kay
// - Created new file.

//============================================================================== 
/*
 *  uBeeDisk - A tool for converting disks/images from one to another with
 *  auto detection options for Microbee disks.
 *  Copyright (C) 2008-2014 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================   
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <ctype.h>
#include <stdint.h>

#ifdef WIN32
#include <windows.h>
#include <winioctl.h>
#include <conio.h>
#include <time.h>
#else
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termio.h>     // used to set terminal modes
#include <termios.h>
#include <sys/types.h>  // various type definitions, like pid_t
#include <sys/time.h>
#endif

#ifdef __linux__
#include <linux/fd.h>
#endif

#ifdef __FreeBSD__
#include <sys/fdcio.h>
#endif

#include "ubeedisk.h"
#include "functions.h"
#include "md5.h"

//==============================================================================
// structures and variables
//==============================================================================
#ifdef WIN32
#else
 static struct timeval tod;
 struct termios term, tOrg;
#endif

#ifdef WIN32
//==============================================================================
// get_key.
//
// Test if a key is ready and return it if there is.
//
//   pass: void
// return: key if ready else -1.
//==============================================================================
int get_key (void)
{
 if (kbhit())
    return getch();
 return -1;
}
#else
//==============================================================================
// get_key() function for standard Unix like systems.  This works in a
// similar way as for the WIN32 method.
//
// See function_init() for the pre-configured variables.
//
// Some keys need to be filtered and other values returned, this may be able
// to be changed by correct configuration.
//
//   pass: void
// return: int key if ready else -1.
//==============================================================================
int get_key (void)
{
 int ch;

 // no delays on reading stdin
 tcsetattr(fileno(stdin), TCSANOW, &term);

 ch = getchar();

 while (getchar() != -1)
    ;
    
 switch (ch)
    {
     case 0x7f : // Unix returns ch == 127 on BS key
        return 8;
     case 0x0a : // Unix returns ch == 10 on ENTER key
        return 13;
     default :
        return ch;
    }
}

//==============================================================================
// printf() wrapper.
//
// Temporarily switches back to default stdin values under unices so that
// output echoes \r to stdout.
//
//   pass: char * fmt,...
// return: int                  value from vsnprintf() function
//==============================================================================
int printfx (char * fmt, ...)
{
 int res;
 va_list ap;
 char buffer[100000];

 va_start(ap, fmt);
 vsnprintf(buffer, sizeof(buffer)-1, fmt, ap);
 buffer[sizeof(buffer)-1] = 0;
 va_end(ap);

 tcsetattr(fileno(stdin), TCSANOW, &tOrg);
 res = fprintf(stdout, "%s", buffer);
 tcsetattr(fileno(stdin), TCSANOW, &term);

 return res;
}

//==============================================================================
// fgets() wrapper.
//
// Temporarily switches back to default stdin values under unices so that
// input is echoed to stdout.
//
//   pass: char *s
//         int size
//         FILE *stream
// return: void
//==============================================================================
char *fgetsx (char *s, int size, FILE *stream)
{
 if (stream == stdin)
    tcsetattr(fileno(stdin), TCSANOW, &tOrg);

#undef fgets
 return fgets(s, size, stream);
#define fgets fgetsx

 if (stream == stdin)
    tcsetattr(fileno(stdin), TCSANOW, &term);
}

#endif

//==============================================================================
// Sleep for milliseconds.
//
//   pass: int ms                       number of milliseconds
// return: void
//==============================================================================
void sleep_ms (int ms)
{
#ifdef WIN32
 Sleep(ms);
#else
 usleep(ms * 1000);
#endif            
}

//==============================================================================
// Get the current clock time in milliseconds
//
//   pass: void
// return: uint64_t                     number of milliseconds
//==============================================================================
uint64_t time_get_ms (void)
{
#ifdef WIN32
 return (uint64_t)clock();
#else
 gettimeofday(&tod, NULL);
 return (((uint64_t)tod.tv_sec * 1000) + ((uint64_t)tod.tv_usec / 1000));
#endif
}

//==============================================================================
// Copy a string to upper case.  Source and destination may be the same.
//
//   pass: char *dest			destination string
//         char *src			source string
// return: void
//==============================================================================
void toupper_string (char *dest, char *src)
{
 int i = 0;
 
 while (src[i])
    {
     dest[i] = toupper(src[i]);
     i++;
    }
 dest[i] = 0;
}

//==============================================================================
// Copy a string to lower case.  Source and destination may be the same.
//
//   pass: char *dest			destination string
//         char *src			source string
// return: void
//==============================================================================
void tolower_string (char *dest, char *src)
{
 int i = 0;
 
 while (src[i])
    {
     dest[i] = tolower(src[i]);
     i++;
    }
 dest[i] = 0;
}

//==============================================================================
// Search an array of strings for the first occurence of the passed search
// string.  The string array must be be terminated by an empty string. 
//
//   pass: char *strg_array	pointer to a string array to be checked
//         char *strg_find	pointer to a string to search for
// return: int			index if found, else -1
//==============================================================================
int string_search (char *strg_array[], char *strg_find)
{
 int i;
 
 for (i=0; strg_array[i][0] != 0; i++)
    {
     if (strcmp(strg_find, strg_array[i]) == 0)
        return i;
    }
 return -1;
}

//==============================================================================
// Convert string to an integer value with error checking.  A -1 value is
// returned if an error is found.
//
// The function uses the strtol() function with the BASE value set to 0 to
// use auto base detection.
//
//   pass: char *s                      string to be converted
// return: int                          integer value
//==============================================================================
int get_integer_value (char *s)
{
 char *check;
 int i;

 if (*s == 0)  // if an empty string
    return -1;

 i = (int)strtol(s, &check, 0);

 if (*check != 0)  // if not end of string then it's an error
    i = -1;

 return i;
}

//==============================================================================
// Parse a string and return the next sub-parameter as a string and an
// integer value.  A pointer to the next parameter position or a NULL is
// returned if the end of string is reached.
//
//   pass: char *s                      string to be parsed
//         int delimiter                sub parameter delimiter
//         char *sps                    sub parameter string returned here
//         int spi                      sub parameter integer returned here
//         int maxlen                   maximum length allowed
// return: char *s                      pointer to next entry
//==============================================================================
char *get_next_parameter (char *s, int delimiter, char *sps, int *spi,
                          int maxlen)
{
 int c = 0;
 char *sps_x = sps;

 if (s == NULL)
    {
     *spi = -1;
     *sps = 0;
     return NULL;
    }

 while ((*s) && (*s != delimiter) && (c++ < maxlen))
    *(sps++) = *(s++);
 *sps = 0;

 *spi = get_integer_value(sps_x);

 if (*s)
    {
     s++;      // get past the delimiter
     if (! *s) // check if delimiter was at end of the line
        return NULL;
     return s;
    }
 else
    return NULL;
}

//==============================================================================
// Return the file name part from the end of a file path.
//
//   pass: char *pathfile		path and file name
//         char *filename		file name part
// return: void
//==============================================================================
void file_name_part (char *pathfile, char *filename)
{
 int i;
 
 // we only want file names, not the complete paths
 i = strlen(pathfile);
 while (i)
    {
     i--;
#ifdef WIN32
     if (pathfile[i] == '\\')
#else
     if (pathfile[i] == '/')
#endif     
        {
         i++; 
         break;
        } 
    }
 strcpy(filename, pathfile+i);
}

//==============================================================================
// Create an MD5 message digest.
//
//   pass: char *filename		file to create MD5 for
//         char *md5			pointer to a 33 byte buffer for MD5
// return: void
//==============================================================================
void create_md5 (char *filename, char *md5)
{
 FILE *fp;
 uint8_t resblock[16];
 int i;

 md5[0] = 0;
 fp = fopen(filename, "rb");
 if (fp)
    {
     if (md5_stream(fp, &resblock) == 0)
        {
         for (i = 0; i < 16; i++)
            sprintf(md5, "%s%02x", md5, resblock[i]);
        } 
     fclose(fp);        
    } 
}

//==============================================================================
// Check operating system version
//
//   pass: void
// return: int					0 if ok, else -1
//==============================================================================
int check_os_version (void)
{
#ifdef WIN32
 // get host system version information
 OSVERSIONINFO osvi;

 ZeroMemory(&osvi, sizeof(OSVERSIONINFO));
 osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
 GetVersionEx(&osvi);

 // make sure we are on running W2000 or higher
 if (osvi.dwMajorVersion >= 5)
    return 0;

 printf("This program must be run on Windows 2000, XP, Vista or higher\n");
 printf("and requires the 'ntwdm' floppy driver to be installed.\n");
 return -1;
 
#else
 return 0;
#endif
}

#ifdef WIN32
//==============================================================================
// Get Windows floppy type
//
//   pass: HANDLE hDrive
// return: int				floppy type (as per Linux)
//==============================================================================
static int get_floppy_type (HANDLE *hDrive)
{
 DISK_GEOMETRY SupportedGeometry[20];

 DWORD ReturnedByteCount;
 int drive_type = 0;
 int geometrys;
 int i;

 if (DeviceIoControl(
    hDrive,
    IOCTL_DISK_GET_MEDIA_TYPES,
    NULL,
    0,
    SupportedGeometry,
    sizeof(SupportedGeometry),
    &ReturnedByteCount,
    NULL
    ))
       geometrys = ReturnedByteCount /  sizeof(DISK_GEOMETRY);
    else
       return 0;

 for (i = 0; i < geometrys; i++)
    {
     switch (SupportedGeometry[i].MediaType)
        {
         // 3.5" drive types
         case F3_1Pt44_512: // 3.5, 1.44MB, 512 bytes/sector
            if (drive_type != 'E')
               drive_type = 'H';
            break;
         case F3_2Pt88_512: // 3.5, 2.88MB, 512 bytes/sector
            drive_type = 'E';
            break;
         case F3_20Pt8_512: // 3.5, 20.8MB, 512 bytes/sector
            break;
         case F3_720_512:   // 3.5, 720KB, 512 bytes/sector
            if ((drive_type != 'H') && (drive_type != 'E'))
               drive_type = 'D';
            break;

         // 5.25" drive types
         case F5_1Pt2_512:  // 5.25, 1.2MB, 512 bytes/sector
            drive_type = 'h';
            break;
         case F5_360_512:   // 5.25, 360KB, 512 bytes/sector
         case F5_320_512:   // 5.25, 320KB, 512 bytes/sector
         case F5_320_1024:  // 5.25, 320KB, 1024 bytes/sector
         case F5_180_512:   // 5.25, 180KB, 512 bytes/sector
         case F5_160_512:   // 5.25, 160KB, 512  bytes/sector
            if (drive_type != 'h')
               drive_type = 'd';
            break;

         case RemovableMedia: // Removable media other than floppy
            break;
         case FixedMedia: // Fixed hard disk media
            break;
         default: // Unknown
            break;
        }
    }

  return drive_type;
}
#endif

//==============================================================================
// Reset drive.
//
// Found that this is necessary when LibDsk is interrupted with a ^C and
// when re-running the application the track position is incorrect, this can
// be seen with a disk scan.  The problem can be corrected by ejecting and
// replacing the disk or by calling ioctl(fd, FDCLRPRM).
//
//   pass: char *xtype				input or output type
//         char *xfile				input or output file
// return: void
//==============================================================================
void reset_drive (char *xtype, char *xfile)
{
#ifdef __linux__
 int fd;
 
 if (strcmp(xtype, "floppy") != 0)
    return;

 fd = open(xfile, O_RDONLY | O_NDELAY);

 if (fd)
    {
     ioctl(fd, FDCLRPRM);
     close(fd);
    }
#endif 
}

//==============================================================================
// Get floppy drive information if xtype is floppy, ntwdm or remote driver.
//
// Returns a letter indicating the drive type.  These values are based on the
// Linux FD FDGETDRVTYP ioctl call.  Windows and FreeBSD will be made to
// return the same corresponding values.
//
// Remote devices (i.e. Floppyio) that support the DRIVETYPE option may be
// used to return the drive type letter.
//
// Linux:
// dnnnn = 5.25 inch double density device files
// hnnnn = 5.25 inch high density device files
// Dnnnn = 3.5 inch double density device files
// Hnnnn = 3.5 inch high density device files
// Ennnn = 3.5 inch extra density device files
//
//   pass: char *xtype				input or output type
//         char *xfile				input or output file
//         DSK_PDRIVER xdrive                   input or output DSK_PDRIVER
// return: int					drive type letter, else 0
//==============================================================================
int get_drive_type (char *xtype, char *xfile, DSK_PDRIVER xdrive)
{
 int res;
 
 // remote drive
 if (! strcmp(xtype, "remote") && xdrive)
    {
     dsk_get_option(xdrive, "DRIVETYPE", &res);
     if (res < 0)
        return 0;
     return res;
    }

#ifdef WIN32
 HANDLE hDrive;
 char device[100];

 if (strcmp(xtype, "ntwdm") != 0)
    return 0;

 snprintf(device, sizeof(device)-1, "\\\\.\\%s", xfile);
 device[sizeof(device)-1] = 0;
 
 hDrive = CreateFile(device, 0, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL);
 if (hDrive == INVALID_HANDLE_VALUE)
    return 0;

 return get_floppy_type(hDrive);
#endif

#ifdef __linux__
 char drivtyp[17];
 int fd;
 
 if (strcmp(xtype, "floppy") != 0)
    return 0;

 drivtyp[0] = 0;    

 fd = open(xfile, O_RDONLY | O_NDELAY);

 if (fd)
    {
     res = ioctl(fd, FDGETDRVTYP, (void *)drivtyp);
     close(fd);
     if (res < 0)
        return 0;
     return drivtyp[0];  // drive type letter
    }
 return 0;
#endif 

#ifdef __FreeBSD__
 enum fd_drivetype fdtype;
 int fd;

 fd = open(xfile, O_RDONLY | O_NDELAY);
 if (fd)
    {
     res = ioctl(fd, FD_GDTYPE, &fdtype);
     close(fd);
     if (res < 0)
        return 0;
     // Diskette drive type, basically the same as stored in RTC on ISA
     // machines (see /sys/isa/rtc.h), but right-shifted by four bits.
     switch (fdtype)
        {
         case FDT_NONE:
            return 0;
         case FDT_360K:
            return 'd';
         case FDT_12M:
            return 'h';
         case FDT_720K:
            return 'D';
         case FDT_144M:
            return 'H';
         case FDT_288M_1:
            return 'E';
         case FDT_288M:
            return 'E';
         default:
            return 0;
        }
    }
 return 0;
#endif

 // All other builds must return 0 if not implementing this function
 return 0;
}

//==============================================================================
// Initialise the functions module.
//
// For programming information: man termios and man termio (older)
//
//   pass: void
// return: void
//==============================================================================
void functions_init (void)
{
#ifndef WIN32
 // get the current stdin values
 tcgetattr(fileno(stdin), &tOrg);

 // this is what is used when requesting single key presses
 // term.c_cc[VMIN] = 0 prevents blocking.
 memset(&term, 0, sizeof(term));
 term.c_lflag &= ~(ECHO | ICANON);
 term.c_cc[VMIN] = 0;
 term.c_cc[VTIME] = 0;
#endif
}

//==============================================================================
// De-initialise the functions module.
//
//   pass: void
// return: void
//==============================================================================
void functions_deinit (void)
{
#ifndef WIN32
 // re-enable the original stdin values
 tcsetattr(fileno(stdin), TCSANOW, &tOrg);
#endif
}
