//******************************************************************************
//*                                 uBeeDisk                                   *
//*                                                                            *
//*      A tool for converting disks/images from one to another with auto      *
//*                  detection options for Microbee disks.                     *
//*                                                                            *
//*                   Copyright (C) 2008-2014 Stewart Kay                      *
//******************************************************************************
//
// This program permits copying and creating disks and images from one to
// another.  The program was created to target Microbee disk formats and to
// provide additional data recovery methods and creating info files.
//
// The program makes use of the 'LibDsk' library and can be found here:
// http://www.seasip.demon.co.uk/Unix/LibDsk/
//
// Parts of this program are based on the tools supplied with LibDsk.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.1 - 11 January 2014, S.J.Kay
// - Fixed format_track() where Microbee DS80 DSK images were not using the
//   correct sector numbers when format_track_range() was being used for
//   partial formatting of a disk (i.e.  when using the --start and --finish
//   options).  DS80 was incorrectly using sectors 1-10 instead of 21-30 for
//   data tracks.  Any formats that make use of xdg.dg_secbase2c to specify
//   data cylinders start point will now use the sector base value held in
//   xdg.dg_secbase2s.
//
// v2.3.0 - 4 June 2013, S.J.Kay
// - Changes to reset_input_drive_and_settings() to exit if not a PC Floppy
//   disk as input type.  This function was causing Remote time-out errors
//   and the code should not be necessary.  Added error reporting and
//   exit(EXIT_FAILURE).
// - Changes to close_files() to report any dsk_close() errors.
// - Removed unused variables xcyl and xhead from disk_speed().
//
// v2.2.0 - 14 May 2012, S.J.Kay
// - Added xdrive parameter to get_drive_type() calls.
// - Removed all the *no_dsk_*[] code and used a new sup_t input_sup
//   structure to replace it.  Added interface_support() function for this.
// - Removed from override_values() RWGAP and FMTGAP check code as will fail
//   first time through and was decided it's not needed anyway?
// - Complete sector ID header value control using extra forceside values.
// - Changes to create_info_file() to report the extra forceside values.
// - Fixed copy_one_disk() format track to only use the disk format's
//   'side1as0' value when the read source type does not support a fully
//   implemented psecid function.
// - Changes made to disk_clean() process to only run if input is a floppy
//   drive.
// - LibDsk call back reporting requires setting verbose level 2 or higher.
// - Added dsk_set_option(idrive, "AUTORATE", 0) for drivers that support
//   this option and support auto rate detection being disabled.
// - Added overwrite_permission() to decide what to do if destination file
//   already exists.
// - Moved pause menu into own function pause_menu() and is also now called
//   from disk_scan(), disk_format() and disk_speed().
// - Disable exiting using ctrl-C, ctrl-\ and ctrl-Z under Unix systems and
//   ctrl-C under windows as can leave floppy in an unstable state.
// - get_drive_type() now only called from open_input_drive() and
//   create_output_drive().
//
// v2.1.0 - 7 October 2011, S.J.Kay
// - Added Abort option to the Pause menu.
// - Added Applix disk skewing to format_track() function.
// - Changes to make reading of all sectors use sector skewing regardless of
//   the buffering mode.  The buffering mode (quiet) is only different in
//   that it does not do any console output and minimal processing between
//   sector reads which might cause sectors to be missed.
// - Changes to read_sector_retry() to report logical and physical sectors.
// - Changes to use logical sector numbers before skew translations.
// - Changes to info_file_entry() to handle skewed sector insertions and to
//   re-instate unread sector flags.
// - Fixed recovery option if drive door is opened when reading a sector and
//   changed the action for the Retry method.
// - Fixed the FLAGS_C5H0S0 format which was broken in v2.0.0.
// - Fixed copy_one_disk() to write out complete info file when copy process
//   is aborted.  This places an 'X' into each unread sector location within
//   the info file.
// - Fixed create_info_file() to report 'X' for unread sector locations.
// - Fixed set_start_finish() to allow override of track start/finish.
// - Changes made to format_set_geometry() parameter passing.
//
// v2.0.0 - 1 October 2011, S.J.Kay
// - Added 'speed' argument to --disk option. This may be used to test and
//   adjust the RPM setting of the spindle motor.
// - Added 'clean' argument to --disk option. This may be used to clean the
//   drive heads using a special cleaning disk.
// - Added track buffering functions read_buffered_track() and
//   write_buffered_track() which greatly speeds up the copy process.
// - Added configurable skewed reading of sectors in track read buffering.
// - Added key detection while copying is in progress, this allows the
//   operator to pause the process.  The operator may then clean the drive
//   heads and continue or switch between Unattended and Interactive modes.
// - Unattended mode is now disabled by default.
// - Changes to read_sector_retry() to allow error handling mode to be
//   changed from the Interactive menu.
// - Changes to read_sector_retry() when using the Interactive menu to
//   Ignore (Skip) all the remaining errors that are encountered for the
//   current disk.
// - Changes to read_sector_retry() prompts.
// - Fixed the disk.oside setting in override_values(), was using wrong
//   condition variable of disk.iside instead of disk.oside which meant the
//   output side was set wrong when disk.iside is not the default unless
//   disk.oside had also been set with an option.
// - disk_scan() on some disks procuded a non recoverable error which
//   prevented all tracks from being scanned successfully.  Added some code
//   to work around this.
// - disk_scan() was incorrectly making use of the --iside option. Changes
//   made so that --scanhead option can specify the start and end heads for
//   scanning.  The default head range is 0-1.
// - disk_scan() now reports LibDsk error when a track can't be scanned.
// - Added read_sector_x() function to replace code used in read_sector().
// - Changes to set_rate_step() to prevent double stepping on auto detect
//   for input/output drives if --idstep=off or --odstep=off are specified.
//   Also now only sets if disk is less than 70 tracks.
// - Changes to read_sector() as some drives error out as it seems there is
//   not enough delay after the head is homed so for the first read error we
//   will do an unreported retry.
// - Added code to override_values() to disallow writing to floppy, remote,
//   etc. if any GAP values are set to 0.
// - Added reset_input_drive_and_settings() function.
// - format_set_geometry(mb_ds84) is now format_set_geometry(xx_default).
// - Removed dg_disktype member from xdg_t structure.
// - Added 'special' member to dg_opts_t structure and added handling for
//   flag '60' disks.
// - Changes to format_track() to improve error reporting.
//
// v1.6.1 - 3 March 2011, S.J.Kay
// - Fixed reading (and writing?) for Honeysoft disks when using a disk
//   driver type that does not support 'dsk_psecid' (i.e 'remote' and
//   FDS.COM on the Microbee computer).  Access past the maximum of 5
//   sectors was occuring on 1024 byte sector tracks, this was fixed by
//   adding some code into read_sector_id().
//
// v1.5.0 - 7 July 2010, S.J.Kay
// - Removed dsk_set_retry(idrive, disk.retries_l1) from copy_one_disk() as
//   this was being overridden in read_sector_id().  The dsk_set_retry()
//   call in read_sector_id() is now used to set the retry value for
//   subsequent sector reads.
// - Added code to override_values() to set the disk.retries_l1 value if no
//   option has been specified.  This value is set to 1 when the input drive
//   type is 'remote' as it was taking too long with all the uneccesary
//   repeated retries on native hardware.  When using the LibDsk driver we
//   tend to see a lot of soft errors so the value needs to be higher.
// - Changed disk_scan() to now use dsk_set_retry(idrive, disk.retries_l1)
//   instead of dsk_set_retry(idrive, 5).  The appropriate value for
//   disk.retries_l1 when scanning is set in an earlier call to
//   override_values().
// - Changes made to fgets() function to use the result as some compilers
//   report warning: declared with attribute warn_unused_result.
//
// v1.4.0 - 18 March 2010, S.J.Kay
// - Fixed problems in read_sector_id() function and other areas by making
//   use of the DSK_ERR_NOTIMPL return value. This value when returned will
//   cause a more standard function call to be used instead. The 'raw' and
//   'remote' types now works when used as an input.  Others types may also
//   be working now if they didn't before.
// - Added verbose level for some console output. The default level is 1
//   which makes all console messages active like previous versions.  If set
//   to 0 using the --verbose option then only error messages, interaction
//   prompts and processes intended to convey information will be output.
// - Added more types to be excluded when creating 'info', and '.err' files. 
//   Only output types that are images will be able to create these files.
//
// v1.3.0 - 12 February 2010, S.J.Kay
// - Added data rate and encoding mode detection and reporting to the
//   disk_scan() process.
// - Changes to disk_scan() process so that scanning can be instructed what
//   side of the disk to read with the --iside option.  If set to default of
//   -1 then both sides are read.
// - Fixed a problem when using batch mode (--count=n) the 'auto_seeked_count'
//   variable was not being set back to 0 for the next copy operation.
// - Fixed disk_copy() to set detection based on disk.itype, not disk.otype.
// - Fixed disk info command (disk_geometry) and moved down.
// - Cleaned up some code in main().
//
// v1.2.0 - 5 February 2010, S.J.Kay
// - Added an unattended operation mode including additional logging
//   information when copying disks.
// - Changes concerning the reading of sectors that was causing many soft
//   retries to show up in the '.info' file and quite likely to have caused
//   problems on some disks because of a too low retry value at the LibDsk
//   level.  Default value of 'disk.retries' (now disk.retries_l2) is now 3
//   instead of 30 and at the LibDsk level (disk.retries_l1) is now 10
//   instead of 1 and can be changed using an option.
// - Disabled the compile of code to test the 'dsk_err_wrt' variable result
//   in copy_one_disk() function as this was failing when writing to disk
//   images depending whether an ignore error response occured earlier when
//   reading a disk.  The actual problem here is still to be established.
// - A new --force option is used to allow the '--disk=format' option to
//   format a disk without user confirmation.
// - Error diagnostics will now be logged to a '.err' file by default
//   except when the --append-error option is specified which will cause
//   the error log to be appended to the 'info' file as before.
//
// v1.1.0 - 10 March 2009, S.J.Kay
// - Transient errors are now processed just like any other disk errors. The
//   error count now also increments when one of these errors occurs.
// - Added error report logging, this is appened to the info file.
//
// v1.0.0 - 8 December 2008, S.J.Kay
// - Initial program created.

//==============================================================================
/*
 *  uBeeDisk - A tool for converting disks/images from one to another with
 *  auto detection options for Microbee disks.
 *  Copyright (C) 2008-2014 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdint.h>
#include <time.h>

#include <libdsk.h>

#ifdef WIN32
#include <windows.h>
#else
#include <signal.h>
#endif

#include "ubeedisk.h"
#include "format.h"
#include "microbee.h"
#include "options.h"
#include "getopt.h"
#include "functions.h"

//==============================================================================
// structures and variables
//==============================================================================
disk_t disk =
{
 .autorate = 1,
 .buffering = 0,
 .count = -1,
 .enter_desc = 1,
 .finish = -1,
 .head_s = 0,
 .head_f = 1,
 .idstep = -1,
#ifdef WIN32
 .ifile = "A:",
#else
 .ifile = "/dev/fd0",
#endif
 .info_file = 1,
 .iside = -1,  // -1 is both sides
 .log = 1,
 .oside = -1,  // -1 is both sides
 .odstep = -1,
 .retries_l1 = -1,
 .retries_l2 = 3,
 .signature = "Not specified",
 .start = -1,
 .unattended = 0,
 .unattended_retry_abort_max = 300,
 .unattended_retry_sector_max = 2,
 .unattended_seeked_max = 10,
 .verbose = 1
};

dg_opts_t dg_opts =
{
 .sidedness = -1,
 .cylinders = -1,
 .heads = -1,
 .sectors = -1,
 .secbase = -1, 
 .secsize = 0,
 .datarate = -1,
 .rwgap = -1,
 .fmtgap = -1,
 .fm = -1,
 .nomulti = -1,
 .noskip = -1,
 .side1as0 = -1, 
 .secbase1s = -1,
 .secbase2s = -1,
 .secbase2c = -1,
 .skew_ofs = -1,
 .skew_val = -1,
 .special = -1
};

xdg_t xdg =
{
 .dg_secbase2c = -1
};

char *data_rates[] =
{
 "hd",      // Data rate for 1.4Mb 3.5"  in 3.5"  drive
 "dd",      // Data rate for 360k  5.25" in 1.2Mb drive
 "sd",      // Data rate for 720k  3.5"  in 3.5"  drive
 "ed"       // Data rate for 2.8Mb 3.5"  in 3.5"  drive
};

char *sidedness[] =
{
 "alt",     // Track n is cylinder (n/heads) head (n%heads)
 "outback", // Tracks go (head 0) 0,1,2,3,...37,38,39, then
            // (head 1) 39,38,37,...,2,1,0
 "outout"   // Tracks go (head 0) 0,1,2,3,...37,38,39, then
            // (head 1) 0,1,2,3,...37,38,39
};

char *noyes[] =
{
 "no",
 "yes"
};

char *sides[] =
{
 "both",
 "0",
 "1"
};

char *forceside_values[] =
{
 "off",
 "on",
 "00",
 "01",
 "10",
 "11",
 ""
};

char *no_error_file[] =
{
 "floppy",
 "ntwdm",
 "remote",
 "rcpmfs",
 ""
};

char *no_info_file[] =
{
 "floppy",
 "ntwdm",
 "remote",
 "rcpmfs",
 ""
};

char *no_desc_otypes[] =
{
 "floppy",
 "ntwdm",
 "remote",
 "rcpmfs",
 ""
};

sup_t input_sup;
sup_t output_sup;

static FILE *infof;
static FILE *errorf;

static int sect_retry_count;
static int sect_retries_tot;
static int sect_errors_tot;

static int auto_retry_abort;
static int auto_retry_sector;
static int auto_seeked_count;
static int auto_cyl_last;
static int auto_head_last;
static int auto_move_head;

char userhome[SSIZE1];
char userhome_confpath[SSIZE1];
static char userhome_path[SSIZE1];
static char error_file[1000];

DSK_GEOMETRY dg;
DSK_GEOMETRY cdg;
DSK_PDRIVER idrive = NULL;

static DSK_PDRIVER odrive = NULL;

static info_t info;

static int overwrite_flag;

static int cyl_start;
static int cyl_finish;

static int skip_all_errors;

static int skew_table[1024];
static uint8_t buf[100000];
static int buffered_cylinder;
static int buffered_head;

static char ofile_name[1000];

// chars = DESC_CHARS + 1 CR + 1 null
static char description[DESC_LINES+1][DESC_CHARS+2]; 

static int diskdesc_line;

extern char *c_argv[];
extern int c_argc;

extern int exitstatus;
extern char *detect_disk_args[];

extern disk_format_t microbee_disk_format[];
extern disk_format_t applix_disk_format[];
extern disk_format_t fm_disk_format[];

//==============================================================================
// Prototypes (static)
//==============================================================================
static void disk_clean (int method);
static int open_input_drive (void);

//==============================================================================
// Report callback function.
//
//   pass: const char *s
// return: void
//==============================================================================
static void report (const char *s)
{
 if (disk.verbose < 2)
    return;
 printf("%s\n", s);
}

//==============================================================================
// Report callback end function.
//
//   pass: void
// return: void
//==============================================================================
static void report_end (void)
{
 if (disk.verbose < 2)
    return;
 printf("\r%-79.79s\r", "");
 fflush(stdout);
}

//==============================================================================
// Ask user for overwrite permission if destination disk file already exists.
//
// If the destination file exists prompt the user for the required action.
// This may be 1 of the following:
//
// [N]o       : do not overwrite this one file.
// [Y]es      : overwrite this one file.
// [A]ll      : overwrite this and all other existing files.
// [S]kip all : do not overwrite this or any other files.
//
// The last 2 are only used if copying more than one disk.
//
// Before calling this function set overwrite_flag=-1 each time the copy
// process is called.
//
//   pass: int action                   1=single file prompts
//                                      2=multiple file prompts
//         char *destfile               path to file to be checked
// return: int                          0=no, 1=yes, 2=skip all
//==============================================================================
static int overwrite_permission (int action, char *destfile)
{
 char user[100];

 // don't prompt the user for action if the --force option was specified
 if (disk.force)
    return 1;

 // if the user selected [A]ll or [S]kip last time
 if (overwrite_flag != -1)
    return overwrite_flag;

 // see if the file can be opened it exists
 FILE *fp = fopen(destfile, "r");

 // return if file does not exist
 if (! fp)
    return 1;
    
 fclose(fp);

 for (;;)
    {
     printf(APPNAME": Destination file already exists. Overwrite \"%s\"\n",
            destfile);
     if (action > 1)
        printf("[N]o [Y]es [A]ll overwrite [S]kip all ?: ");
     else
        printf("[N]o [Y]es ?: ");
     fflush(stdout);
 
     if (fgets(user, sizeof(user)-2, stdin))
        {} // ignore result 

     switch (toupper(user[0]))
        {
         case 'N' : // no overwrite
            return 0;
         case 'Y' : // overwrite 1
            return 1;
         case 'A' : // overwrite all
            if (action > 1)
               {
                overwrite_flag = 1;
                return 1;
               }
            break;
         case 'S' : // skip all
            if (action > 1)
               {
                overwrite_flag = 2;
                return 2;
               }
            break;
        }
    }
}

//==============================================================================
// Check if user is requesting the pause menu and act on selections.
//
// Allow the operator to pause, abort, clean heads and switch between
// unattended and interactive modes depending on requested options.
//
//   pass: int options                  0=+RACM, 1=+RAC-M, 2=+RA-CM
// return: int                          result value
//==============================================================================
static int pause_menu (int options)
{
 char user[100];

 if (get_key() == -1)
    return DSK_ERR_OK;
    
 user[0] = 0;
 printf("\n");         
 while (toupper(user[0]) != 'R')
    {
     printf("\nPAUSE MENU\n");
     printf("----------\n");

     switch (options)
        {
         case 0 : 
            if (disk.unattended)
               printf("[R]esume [A]bort [C]lean [M]ode set to interactive ?: ");
            else
               printf("[R]esume [A]bort [C]lean [M]ode set to unattended ?: ");
            break;
         case 1 : 
            printf("[R]esume [A]bort [C]lean ?: ");
            break;
         case 2 : 
            printf("[R]esume [A]bort ?: ");
            break;
        }
         
     fflush(stdout);

     if (fgets(user, sizeof(user)-2, stdin))
        {} // ignore result 

     switch (toupper(user[0]))
        {
         case 'A' : // abort the copy (not an error from this menu)
            return DSK_ERR_ABORT;                 
         case 'C' : // clean drive heads with special disk
            if (options > 1)
               break;
            disk_clean(0);
            break;
         case 'R' : // resume
            break;
         case 'M' : // switch between unattended and interactive mode
            if (options != 0)
               break;
            disk.unattended = ! disk.unattended;
            if (disk.unattended)
               printf("Unattended mode set!\n");
            else
               printf("Interactive mode set!\n");
            break;
         default :
            break;
        }
    }

 printf("\n");
 return DSK_ERR_OK;  
}

//==============================================================================
// Log a read error.
//
// Logs read errors if 'info' files are enabled and the log level is not 0.
// At completion a '.err' file will be created or may be optionally appended
// to the '.info' file if requested.
//
// The error file is only created if there were errors encountered.
//
//   pass: dsk_err_t dsk_err		LibDsk disk error
//         int tryx                     try number
//         dsk_pcyl_t cyl
//         dsk_phead_t head
//         int lsect                    logical sector number
// return: void
//==============================================================================
static void log_read_error (dsk_err_t dsk_err, int tryx, dsk_pcyl_t cyl,
                            dsk_phead_t head, int lsect)
{
 if ((! disk.info_file) || (! disk.log) ||
    (string_search(no_error_file, disk.otype) != -1))
    return;

 if (! errorf)
    {
     // create the error log file
     snprintf(error_file, sizeof(error_file), "%s.err", ofile_name);
     error_file[sizeof(error_file)-1] = 0;

     errorf = fopen(error_file, "w+");
     if (errorf == NULL)
        {
         printf(APPNAME": unable to create error file: %s\n", error_file);
         return;
        }
     else
        {
         fprintf(errorf, "LOG OF READ ERRORS\n");
         fprintf(errorf, "------------------\n");

         fprintf(errorf,
 "This log contains errors that were encountered during the imaging process.\n"
 "Read errors may have been corrected on further retries.\n"
 "\n"
 "The sector count at 'Sector errors' and the 'SECTOR STATUS MAP' in the\n"
 "'info' file should be used to determine what bad sectors the disk (image)\n"
 "may have.\n"
 "\n"
 "The frequency of log entries recorded here is determined by the logging\n"
 "level option '--log=n'.\n"
 "\n");
        }
    }

 if (errorf)
    {
     fprintf(errorf,
     "Try #: %02d/%02d Cylinder: %02d Head: %d Sector: %03d Error: (%d) %s\n",
     tryx, disk.retries_l2, cyl, head, skew_table[lsect] + dg.dg_secbase,
     dsk_err, dsk_strerror(dsk_err));
    } 
}

//==============================================================================
// Format a disk track.
//
// Most disks use a sequential order on the disk tracks but depending on
// flags for the format this can be varied.
//
//   pass: dsk_pcyl_t cyl		cylinder number
//         dsk_phead_t head		physical side of disk
//         int side			side ID value
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
static dsk_err_t format_track (dsk_pcyl_t cyl, dsk_phead_t head, int side)
{
 dsk_err_t dsk_err = DSK_ERR_OK;
 
 int applix[2][5] =
 {
  {4,2,5,3,1},
  {1,4,2,5,3}
 };

 DSK_FORMAT format[256];
 int i;
    
 for (i = 0; i < dg.dg_sectors; i++)
    {
     format[i].fmt_cylinder = cyl;
     format[i].fmt_head = side;

     if (xdg.dg_special & FLAGS_APPLIX)
        format[i].fmt_sector = applix[head][i];
     else
        if ((xdg.dg_secbase2c != -1) && (cyl >= xdg.dg_secbase2c))
            format[i].fmt_sector = xdg.dg_secbase2s + i;
         else   
            format[i].fmt_sector = xdg.dg_secbase1s + i;

     format[i].fmt_secsize = dg.dg_secsize;
    }

 // avoid dsk_pformat() for formats that do not support the function
 if (! output_sup.pformat)
    return DSK_ERR_OK;

 dsk_err = dsk_pformat(odrive, &dg, cyl, head, format, 0xE5);

 // ignore formatting if the driver does not support the format function. 
 if (dsk_err == DSK_ERR_NOTIMPL)
    return DSK_ERR_OK;
 
 if (dsk_err != DSK_ERR_OK)
    {
     printf("\n"APPNAME
     ": format_track() - Cylinder: %02d Head: %d Side: %03d Error: (%d) %s\n",
     cyl, head, side, dsk_err, dsk_strerror(dsk_err));
    }

 return dsk_err;   
}

//==============================================================================
// Format a range of disk tracks.
//
//   pass: int start			cylinder start number
//         int finish			cylinder finish number 
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
static dsk_err_t format_track_range (int start, int finish)
{
 dsk_err_t dsk_err = DSK_ERR_OK;
 dsk_pcyl_t cyl;
 dsk_phead_t head = 0;

 // format tracks if start > cylinder 0
 if (disk.noformat)
    return DSK_ERR_OK;

 if (disk.verbose)
    printf("Formatting cylinders: %d to %d.\n", start, finish);

 for (cyl = start; cyl <= finish; cyl++)
    {
     for (head = 0; head < dg.dg_heads; head++)
        {
         if (disk.verbose)
            {
             printf("\rCylinder: %02d/%02d Head: %d/%d",
                    cyl, dg.dg_cylinders-1, head, dg.dg_heads-1);
             fflush(stdout);
            } 
         dsk_err = format_track(cyl, head, head);
        }
    }

 if (dsk_err != DSK_ERR_OK)
    {
     printf("\nWrite error: %s\n", dsk_strerror(dsk_err));
     printf("Aborting this disk copy.\n");
     return dsk_err;
    }

 if (disk.verbose)
    printf("\n");

 return dsk_err;
}

//==============================================================================
// Generate skew data for reading sectors based on the parameters passed.
//
//   pass: int skew                     skewing value
//         int start                    first logical sector in table
//         int sectors                  number of sectors per track
// return: void
//==============================================================================
static void create_skew_table (int skew, int start, int sectors)
{
 int i;
 int k;
 int j = start;
 
 for (i = 0; i < sectors; ++i)
    {
     for (;;)
        {
         for (k = 0; k < i && skew_table[k] != j; ++k)
            ;
    
         if (k < i)
            j = (j + 1) % sectors;
         else
            break;
        }
    
     skew_table[i] = j;
     
     j = (j + skew) % sectors;
    }
}

//==============================================================================
// Read a complete track of data into a global buffer.
//
// All tracks are buffered into memory regardless of the --buffering mode
// setting.  The buffering mode (quiet) is only different in that it does
// not do any console output and minimal processing between sector reads
// which might cause sectors to be missed.
//
// Disk reading time can be greatly reduced by reading a complete track into
// memory then writing it out.  It was found that reading every alternate
// sector (skew=2) starting with the 2nd sector on disks that have a
// sequential format works well.
//
// If the track read fails then the calling process should manage the
// reading of all the sectors on the track concerned.
//
// To see the skew values use --verbose=2 on the command line.
//
// Microbee DS40 examples:
//
// disk.skew_ofs = 1 and disk.skew_val = 2 (default)
// 2  4  6  8 10  3  5  7  9  1
//
// disk.skew_ofs = 0 and disk.skew_val = 2
// 1  3  5  7  9  2  4  6  8 10
//
// disk.skew_ofs = 0 and disk.skew_val = 1
// 1  2  3  4  5  6  7  8  9 10
//
//
//   pass: dsk_pcyl_t cyl		physical drive cylinder number
//         dsk_pcyl_t xcyl		ID cylinder value
//         dsk_phead_t head		physical drive side
//         dsk_phead_t xhead		ID side value
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
static dsk_err_t read_buffered_track (dsk_pcyl_t cyl, dsk_pcyl_t xcyl,
                                      dsk_phead_t head, dsk_phead_t xhead)
{
 // enable this to speed up testing and use --verbose=2 --finish=0 --heads=1
 // on the command line.
 uint8_t *p;

 int psect;
 int i;
 int skip_read;
 dsk_err_t implemented = DSK_ERR_OK;
 dsk_err_t dsk_err = DSK_ERR_OK;

 // avoid dsk_xread() for types that do not support the function
 if (! input_sup.xread)
    implemented = DSK_ERR_NOTIMPL;

 // read in the complete track without any console IO (unless --verbose > 1)
 for (i = 0; i < dg.dg_sectors; i++)
    {
     dsk_err = implemented;

     // get the skewed physical sector number
     psect = skew_table[i] + dg.dg_secbase;

     // calculate the buffer pointer location
     p = buf + dg.dg_secsize * (psect - dg.dg_secbase);

     // C5H0S0 disk format ? (sectors on C5, H0 are 0,1,2,3,4,5,6,7,8,9)
     if ((skip_read = ((xdg.dg_special & FLAGS_C5H0S0) && cyl == 5 &&
        head == 0 && psect == 0)))
        {
         memset(p, 0, dg.dg_secsize);
         dsk_err = DSK_ERR_OK;
        }
     
     if (! skip_read)
        {
         if (dsk_err != DSK_ERR_NOTIMPL)
            dsk_err = dsk_xread(idrive, &dg, p, cyl, head, xcyl, xhead, psect,
                                dg.dg_secsize, NULL);
         if (dsk_err == DSK_ERR_NOTIMPL)
            dsk_err = dsk_pread(idrive, &dg, p, cyl, head, psect);

         // abort the track read if an error (will use sector by sector)
         if (dsk_err != DSK_ERR_OK)
            break;
        }
    }

 // if error save the cylinder and head as negative values before returning
 if (dsk_err != DSK_ERR_OK)
    {
     buffered_cylinder = -2;  // flag it to use non buffered reads
     return dsk_err;
    } 

 // save the buffered locations
 buffered_cylinder = cyl;
 buffered_head = head;
 
 return dsk_err;
}

//==============================================================================
// Write a buffered track.
//
//   pass: dsk_pcyl_t cyl		physical drive cylinder number
//         dsk_pcyl_t xcyl		ID cylinder value
//         dsk_phead_t head		physical drive side
//         dsk_phead_t xhead		ID side value
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
static dsk_err_t write_buffered_track (dsk_pcyl_t cyl, dsk_pcyl_t xcyl,
                                       dsk_phead_t head, dsk_phead_t xhead)
{
 uint8_t *p;
 int i;
 dsk_pcyl_t use_cyl;
 dsk_phead_t use_head = xhead;
 dsk_err_t dsk_err = DSK_ERR_OK;

#ifndef DEBUG_BUFFERING 
 for (i = 0; i < dg.dg_sectors; i++)
    {
     // calculate the buffer pointer location
     p = buf + dg.dg_secsize * i;

     // write the sector to the output file
     dg.dg_datarate = xdg.dg_odatarate; // set data rate for output

     // avoid dsk_xwrite() for types that do not support the function
     if (! output_sup.xwrite)
        dsk_err = DSK_ERR_NOTIMPL;
     else
        {
         use_cyl = cyl;

         switch (disk.forceside)
            {
             case 0 : // off
                use_cyl = xcyl;
                use_head = xhead;
                break;
             case 1 : // on
                use_head = head;
                break;
             case 2 : // 00
                use_head = 0;
                break;
             case 3 : // 01
                if (head == 0)
                   use_head = 0;
                else
                   use_head = 1;
                break;
             case 4 : // 10
                if (head == 0)
                   use_head = 1;
                else
                   use_head = 0;
                break;
             case 5 : // 11
                use_head = 1;
                break;
            }        

        dsk_err = dsk_xwrite(odrive, &dg, p, cyl, head, 
        use_cyl, use_head, dg.dg_secbase + i, dg.dg_secsize, 0);
        }
     
     if (dsk_err == DSK_ERR_NOTIMPL)
        dsk_err = dsk_pwrite(odrive, &dg, p, cyl, head,
        dg.dg_secbase + i);

     if (dsk_err != DSK_ERR_OK)
        return dsk_err;
    }
#endif
 
 return dsk_err;
}

//==============================================================================
// Read sector ID to find the side ID.
//
// If the 'raw' or any other drivers not supporting the dsk_psecid()
// function then the side ID will be returned with the passed head value.
//
// The passed head value is also returned if there is an error reading a
// sector ID.  This is allowed to return as OK and relies on further sector
// reading to report an error.
//
// A high LibDsk retry level is set here to help make sure there is plenty
// of time to find an ID.  A constant is used but is changed for the actual
// sector reads that may follow.
//
//   pass: dsk_pcyl_t cyl		cylinder number
//         dsk_phead_t head		physical side of disk
//         dsk_phead_t *res_head	head result
//         int *res_size		sector size result
// return: dsk_err_t			DSK_ERR_OK
//==============================================================================
static int read_sector_id (dsk_pcyl_t cyl, dsk_phead_t head,
                           dsk_phead_t *res_head, int *res_size)
{
 dsk_err_t dsk_err = DSK_ERR_OK;
 DSK_FORMAT result;

 dg.dg_datarate = xdg.dg_idatarate; // set data rate for input

 dsk_set_retry(idrive, disk.retries_l1); // set retries for the input drive

 // avoid dsk_psecid() for types that do not support the function
 if (! input_sup.psecid)
    dsk_err = DSK_ERR_NOTIMPL;  // this will cause the values below to be used
 else
    // get the side information from first available sector header ID
    dsk_err = dsk_psecid(idrive, &dg, cyl, head, &result);

 if (dsk_err == DSK_ERR_OK)
    {
     *res_head = result.fmt_head;
     *res_size = result.fmt_secsize;
    }
 else // else if DSK_ERR_NOTIMPL or read error
    {
     // Honeysoft format ?
     if (xdg.dg_special & FLAGS_HSOFT)
        {
         if (cyl >= 5)
            *res_size = 1024;
         else
            *res_size = 512;
        }
     else
        *res_size = cdg.dg_secsize;  // original geometry sector size
     *res_head = head;
    }

 return DSK_ERR_OK;  // we always return OK here.
}

//==============================================================================
// Read a sector from the input drive using the appropriate function into
// the global buffer.
//
//   pass: dsk_pcyl_t cyl		physical drive cylinder number
//         dsk_pcyl_t xcyl		ID cylinder value
//         dsk_phead_t head		physical drive side
//         dsk_phead_t xhead		ID side value
//         int lsect			logical sector number
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
static dsk_err_t read_sector_x (dsk_pcyl_t cyl, dsk_pcyl_t xcyl,
                                dsk_phead_t head, dsk_phead_t xhead,
                                int lsect)
{
 uint8_t *p;
 int psect;
 dsk_err_t dsk_err = DSK_ERR_OK;

 // read in a one buffered track without console IO and other processing if
 // buffering mode is set to quiet.
 if (disk.buffering == 1)
    {
     // if this was recorded as bad during a previous track read then skip
     // trying to buffer using quiet mode for the same error prone track and
     // do it using the normal mode instead.
     if (buffered_cylinder != -2)
        {
         // read in the track if we don't alread have it buffered
         if (buffered_cylinder != cyl || buffered_head != head)
            dsk_err = read_buffered_track(cyl, xcyl, head, xhead);
     
         // if the track was read in ok we are done!
         if (dsk_err == DSK_ERR_OK)
            return dsk_err;
        }
    }    

 // get the skewed physical sector number
 psect = skew_table[lsect] + dg.dg_secbase;

 // calculate the buffer pointer location
 p = buf + dg.dg_secsize * (psect - dg.dg_secbase);

 // C5H0S0 disk format ?
 if ((xdg.dg_special & FLAGS_C5H0S0) && cyl == 5 && head == 0 && psect == 0)
    {
     memset(p, 0, dg.dg_secsize);
     return DSK_ERR_OK;
    }

 // avoid dsk_xread() for types that do not support the function
 if (! input_sup.xread)
    dsk_err = DSK_ERR_NOTIMPL;
 else
    dsk_err = dsk_xread(idrive, &dg, p, cyl, head, xcyl, xhead, psect,
    dg.dg_secsize, NULL);
 
 if (dsk_err == DSK_ERR_NOTIMPL)
    dsk_err = dsk_pread(idrive, &dg, p, cyl, head, psect);
 
 return dsk_err;
}

//==============================================================================
// Read a sector from the input drive.
//
// Uses side checking and first level automatic retrying on read errors
// after the side information of sector IDs is found.
//
//   pass: dsk_pcyl_t cyl		cylinder number
//         dsk_phead_t head		physical side of disk
//         dsk_phead_t  xhead		side ID value
//         int lsect			logical sector number
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
static dsk_err_t read_sector (dsk_pcyl_t cyl, dsk_phead_t head,
                              dsk_phead_t xhead, int lsect)
{
 dsk_err_t dsk_err = DSK_ERR_OK;
 int count;
 int tryx;
 
 count = disk.retries_l2;  // this is level 2 (up from LibDsk retries)
 tryx = 0;

 dg.dg_datarate = xdg.dg_idatarate; // set data rate for input

 for (;;)
    {
     tryx++;

     for (;;)
        {
         // read the sector into the global buffer
         dsk_err = read_sector_x(cyl, cyl, head, xhead, lsect);

         // some drives error out as it seems there is not enough delay
         // after the head is homed so for the first read error we will do
         // an unreported retry. (i.e Tandon TM848-02 8") 
         if (! disk.first_read)
            break;
         disk.first_read = 0;
         if (dsk_err == DSK_ERR_OK)
            break;

         // don't use reset_input_drive_and_settings() here! instead just
         // use a delay or the head will try to home constantly.
         sleep_ms(3000);
        }

     if (dsk_err == DSK_ERR_OK)
        return dsk_err;

     // if logging all read errors (good for drive diagnostics)
     if (disk.log == 2)
        log_read_error(dsk_err, tryx, cyl, head, lsect);

     sect_retry_count++;
     sect_retries_tot++;

     if (--count == 0)
        {
         printf("\n"APPNAME": read_sector() - %s\n", dsk_strerror(dsk_err));
         // if logging only errors when the retry count expires
         if (disk.log == 1)
            log_read_error(dsk_err, tryx, cyl, head, lsect);
         return dsk_err;
        }
    }    
}

//==============================================================================
// Read a sector from the input drive using recovery methods on errors.
//
// In interactive mode (--unattended=off) There is a choice of 6 options
// the operator may select when an error occurs:
//
// 1. reTry - Will try another read but does not seek to track 0 first.
// 2. Retry - Will seek to cylinder 0 and read a sector ID before retrying.
//    If the cylinder is already on 0 then cylinder 9 is used instead.
// 3. Ignore - An error is recorded and exits with no error.  The data is
//    likely to be corrupt.
// 4. Skip - Same as Ignore but for all remaining errors on current disk.
// 5. Abort - An error is recorded and exits with a disk error abort value.
// 6. Clean - Clean the drive heads using a special cleaning disk.
// 7. Mode - Toggle between Unattended and Interactive mode.
//
//   pass: dsk_pcyl_t cyl		cylinder number
//         dsk_phead_t head		physical side of disk
//         dsk_phead_t  xhead		side ID value
//         int lsect			logical sector number
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
static dsk_err_t read_sector_retry (dsk_pcyl_t cyl, dsk_phead_t head,
                                    dsk_phead_t xhead, int lsect)
{
 dsk_err_t dsk_err = DSK_ERR_OK;
 DSK_FORMAT result;

 char user[100];
 sect_retry_count = 0;
 auto_retry_sector = 0;  // reset for the next sector
 
 for (;;)
    {
     // allow the operator to pause, resume, abort, clean and switch between
     // unattended and interactive modes
     if (pause_menu(0) == DSK_ERR_ABORT)
        {
         sect_retry_count = -2;
         return DSK_ERR_ABORT;
        }
     
     dsk_err = DSK_ERR_OK;
     
     if (disk.verbose)
        {
         printf("\rCylinder: %02d/%02d Head: %d/%d"
                " Sector: (log) %03d/%03d, (phy) %03d/%03d",
                cyl, dg.dg_cylinders - 1, head, dg.dg_heads - 1,
                lsect, dg.dg_sectors - 1,
                skew_table[lsect] + dg.dg_secbase,
                dg.dg_sectors + dg.dg_secbase - 1);
         fflush(stdout);
        }

     dsk_err = read_sector(cyl, head, xhead, lsect);

     if (disk.verbose)
        {
         if (dsk_err != DSK_ERR_OK)
            printf("\n");
        }    

     if (dsk_err == DSK_ERR_OK)
        return dsk_err;

     if (disk.unattended)
        {
         // If using the un-attended mode then errors are handled using a
         // pre-programmed series of actions.
         //
         // The 'auto_retry_sector' counts the number of retries for a given
         // sector and is reset to 0 for each new sector.  When this reaches
         // the maximum allowed value the 'auto_retry_abort' counter is
         // incremented and if that reaches the maximum allowed value then
         // the read process is aborted.
         //
         // Each time an error occurs a check will be made to see if a retry
         // using a seek track 0 should be initiated first or simply just
         // retry reading the same sector without any track 0 seek.  Once
         // the maximum number of track 0 seeks have been used up only
         // retries without seeking T0 are issued. (This is to reduce wear
         // on disks)
         
         if ((auto_cyl_last != cyl) || (auto_head_last != head))
            {
             auto_move_head = 1;
             auto_cyl_last = cyl;
             auto_head_last = head;
             auto_retry_sector = 0;
            } 
         
         // once the retries are used up for the current sector read
         // increase the abort counter and if the maximum value is reached
         // abort.

         if (++auto_retry_sector > disk.unattended_retry_sector_max)
            {
             if (++auto_retry_abort > disk.unattended_retry_abort_max)
                user[0] = 'A';
             else
                user[0] = 'I';
            }
         else   
            {
             // decide if the head will be moved first before retrying

             if ((auto_move_head) &&
                (auto_seeked_count < disk.unattended_seeked_max))
                {
                 user[0] = 'R';
                 auto_move_head = 0;
                 auto_seeked_count++;
                }
             else
                user[0] = 'T';
            } 

         printf("Unattended mode:\n");
         printf("re[T]ry (no home) [R]etry (seek home) [I]gnore [A]bort"
                " ---> '%c'\n", user[0]);
         if ((user[0] != 'I') && (user[0] != 'A'))
            printf("Retries for disk: %d/%d  Retries for sector:"
                   " %d/%d  Track #0 retry seeks: %d/%d\n\n",
            auto_retry_abort,  disk.unattended_retry_abort_max,
            auto_retry_sector, disk.unattended_retry_sector_max,
            auto_seeked_count, disk.unattended_seeked_max);
         else   
            printf("Error in read operation, ignoring last sector and"
                   " recording as an error!\n\n");
        }
     else
        {
         if (skip_all_errors)
            user[0] = 'I';
         else
            {
             // Operator to decide what to do. This also allows the unattended
             // or Interactive mode to be set or the drive heads to be cleaned.
             user[0] = 0;
             while ((toupper(user[0]) != 'T') && (toupper(user[0]) != 'R') &&
                    (toupper(user[0]) != 'I') && (toupper(user[0]) != 'A') &&
                    (toupper(user[0]) != 'S'))
                {
                 printf("INTERACTIVE ERROR MENU\n");
                 printf("----------------------\n");
                 printf("re[T]ry (no home) [R]etry (home) [I]gnore [S]kip"
                        " [A]bort [C]lean [M]ode ?: ");
                 fflush(stdout);
                 if (fgets(user, sizeof(user)-2, stdin))
                    {} // ignore result
                 switch (toupper(user[0]))
                    {
                     case 'C' : // clean drive heads with special disk
                        disk_clean(0);
                        printf("\n");
                        break;
                     case 'M' : // switch to unattended or interactive mode                 
                        disk.unattended = ! disk.unattended;
                        if (disk.unattended)
                           printf("Unattended mode set!\n");
                        else
                           printf("Interactive mode set!\n");
                        printf("\n");
                        break;
                     default :
                        printf("\n");
                        break;
                    }
                }
            }
        }

     switch (toupper(user[0]))
        {
         case 'T' : // retry but don't move head
                    break;
         case 'R' : // retry after moving the head first
                    // Need to do the following if drive door was opened
                    // when trying to read on some drives.
                    dsk_psecid(idrive, &dg, 0, 0, &result);
                    reset_input_drive_and_settings();

                    // if we were on track 0 we move the head to T0 first
                    // then back again.
                    if (cyl != 0)
                       break;
                    dsk_psecid(idrive, &dg, 1, 0, &result);
                    dsk_psecid(idrive, &dg, 0, 0, &result);
                    break;
         case 'S' : // All future errors for this disk will be ignored.
                    // First time we fall through to case 'I'.
                    skip_all_errors = 1;
         case 'I' : // mark as an error but we return an OK to keep going
                    sect_errors_tot++;
                    sect_retry_count = -1;
                    
                    // Need to do the following if drive door was opened
                    // when trying to read on some drives.
                    dsk_psecid(idrive, &dg, 0, 0, &result);
                    reset_input_drive_and_settings();
                    return DSK_ERR_OK;    
         case 'A' : // abort the copy with an error
                    sect_errors_tot++;
                    sect_retry_count = -1;
                    return DSK_ERR_ABORT; 
        }
    }
}

//==============================================================================
// Initialise the info file values.
//
// Initialises some variables.  The entire info file is created in the
// create_info_file() function.
//
//   pass: void
// return: void
//==============================================================================
static void init_info_file (void)
{
 info.cyl_last = -1;
 info.head_last = -1;
 info.pos_trk = 0;
 info.buf[0] = -1;
}

//==============================================================================
// Create info file
//
//   pass: void
// return: void
//==============================================================================
static void create_info_file (void)
{
 char *month_names[] =
 {
  "January",
  "February",
  "March",
  "April",
  "May",
  "June",
  "July",
  "August",
  "September",
  "October",
  "November",
  "December"
 };

 typedef struct tm tm_t;

 time_t result;
 tm_t resultp;

 char temp_str[100];
 char date_time[100];
 
 int i;
 int x;
 int retries;

 int secbase_last = -1;
 int sectors_last = -1;
 int secsize_last = -1;

 char *res;
 char inpf[1000];
 char outf[1000];
 char md5[33];

 char info_file[1000];

 // check if output type is a disk or other non-image and skip the info file
 // if it is.
 if ((! disk.info_file) || (string_search(no_info_file, disk.otype) != -1))
    return;

 // create the info file
 snprintf(info_file, sizeof(info_file), "%s.info", ofile_name);
 info_file[sizeof(info_file)-1] = 0;

 infof = fopen(info_file, "w");
 if (infof == NULL)
    {
     printf(APPNAME": unable to create info file: %s\n", info_file);
     return;
    }

 // create a date and time stamp
 time(&result); 
#ifdef WIN32
 memcpy(&resultp, localtime(&result), sizeof(resultp));
#else
 localtime_r(&result, &resultp);
#endif 
 if (resultp.tm_sec > 59)		// 2 second leap-seconds ignore
    resultp.tm_sec = 59;

 sprintf(date_time, "%d %s %d  %02d:%02d:%02d", resultp.tm_mday,
 month_names[resultp.tm_mon],
 resultp.tm_year+1900, resultp.tm_hour, resultp.tm_min, resultp.tm_sec);
 
 // compute an MD5 for the image created
 create_md5(ofile_name, md5);
 
 // extract the file names from the paths
 file_name_part(disk.ifile, inpf);
 file_name_part(ofile_name, outf);

 if (! infof)
    return;
    
 fprintf(infof, "DISK/IMAGE INFORMATION\n");
 fprintf(infof, "----------------------\n");
 fprintf(infof, "File output        %s\n", outf);
 fprintf(infof, "File output MD5    %s\n", md5);
 fprintf(infof, "Disk name          %s\n", xdg.dg_format_name);
 fprintf(infof, "Disk name desc     %s\n", xdg.dg_format_desc);
 fprintf(infof, "Archived by        %s\n", disk.signature);
 fprintf(infof, "Creation date      %s\n\n", date_time);
 
 fprintf(infof, "DISK DESCRIPTION\n");
 fprintf(infof, "----------------\n");
 i = 0;
 for (;;)
    {
     if ((description[i][0]) && (strcmp(".\n", description[i]) != 0))
        fprintf(infof, "%s", description[i++]);
     else
        break;
    }
 fprintf(infof, "\n");   

 fprintf(infof, "PROGRAM PARAMETERS\n");
 fprintf(infof, "------------------\n");          
 fprintf(infof, "Image program      "TITLESTRING"\n");
 fprintf(infof, "Detection req      %s\n", detect_disk_args[disk.detect]);
 fprintf(infof, "Disk format req    %s\n", disk.format);
 fprintf(infof, "File input         %s\n", inpf);
 fprintf(infof, "Input type         %s\n", disk.itype);
 fprintf(infof, "Output type        %s\n", disk.otype);
 fprintf(infof, "Input dstep        %s\n", noyes[disk.idstep_used]);
 fprintf(infof, "Output dstep       %s\n", noyes[disk.odstep_used]);
 fprintf(infof, "Input side         %s\n", sides[disk.iside+1]);
 fprintf(infof, "Output side        %s\n", sides[disk.oside+1]);
 fprintf(infof, "Start cylinder     %d\n", cyl_start);
 fprintf(infof, "Finish cylinder    %d\n", cyl_finish);
 fprintf(infof, "Force side         %s\n", forceside_values[disk.forceside]);
 fprintf(infof, "Input comp         %s\n", disk.incomp);
 fprintf(infof, "Output comp        %s\n", disk.outcomp);
 fprintf(infof, "Retries lev1 (max) %d\n", disk.retries_l1);
 fprintf(infof, "Retries lev2 (max) %d\n", disk.retries_l2);
 fprintf(infof, "Logging level      %d\n", disk.log);
 fprintf(infof, "Input data rate    %s\n", data_rates[xdg.dg_idatarate]);
 fprintf(infof, "Output data rate   %s\n", data_rates[xdg.dg_odatarate]);
 fprintf(infof, "Read buffering     %s\n", noyes[disk.buffering]);

 fprintf(infof, "\n");
 fprintf(infof, "UNATTENDED MODE SETTINGS\n");
 fprintf(infof, "------------------------\n"); 
 fprintf(infof, "Used unattended    %s\n", noyes[disk.unattended]);
 fprintf(infof, "Sector ret/abort   %d\n", disk.unattended_retry_abort_max); 
 fprintf(infof, "Retries/sector max %d\n", disk.unattended_retry_sector_max); 
 fprintf(infof, "Error seeks max    %d\n", disk.unattended_seeked_max); 

 fprintf(infof, "\n");
 fprintf(infof, "PHYSICAL ATTRIBUTES\n");
 fprintf(infof, "-------------------\n");
 fprintf(infof, "Sidedness          %s\n", sidedness[dg.dg_sidedness]);
 fprintf(infof, "Data rate          %s\n", data_rates[xdg.dg_odatarate]);
 fprintf(infof, "Cylinders          %d\n", dg.dg_cylinders);
 fprintf(infof, "Heads              %d\n", dg.dg_heads);
 if (xdg.dg_special & FLAGS_HSOFT)
    {
     fprintf(infof, "Sectors/track      10x512 and 5x1024\n");
     fprintf(infof, "Sector size        512 and 1024\n");
    }
 else
    {
     fprintf(infof, "Sectors/track      %d\n", dg.dg_sectors);
     fprintf(infof, "Sector size        %d\n", dg.dg_secsize);
    }
 fprintf(infof, "Sector base #1     %d\n", xdg.dg_secbase1s);
 fprintf(infof, "Sector base #2     %d\n", xdg.dg_secbase2s);
 fprintf(infof, "Read skew          %d\n", xdg.dg_skew_val);
 fprintf(infof, "Read skew offset   %d\n", xdg.dg_skew_ofs);

 fprintf(infof, "\n");
 fprintf(infof, "DISK COPY RESULTS\n");     
 fprintf(infof, "-----------------\n");
 fprintf(infof, "Sector errors      %d\n", sect_errors_tot);
 fprintf(infof, "Sector retries     %d\n", sect_retries_tot);     


 fprintf(infof, "\n");
 fprintf(infof, "SECTOR STATUS MAP\n");
 fprintf(infof, "-----------------\n");

 fprintf(infof,
"A sector position containing a '.' indicates the sector read correctly first\n"
"time. An 'X' indicates the sector was not read (terminated early). An 'ERR'\n"
"indicates the sector had a read error and a value from '1-999' indicates the\n"
"number of level 2 retries that were needed to successfully read the sector."
"\n\n");

 i = 0;
 while (info.buf[i] != -1)
    {
     if ((info.buf[i] != secbase_last) || (info.buf[i+1] != sectors_last) ||
        (info.buf[i+2] != secsize_last))
        {
         secbase_last = info.buf[i];
         sectors_last = info.buf[i+1];
         secsize_last = info.buf[i+2];
         sprintf(temp_str, "(%dx%d)", sectors_last, secsize_last);
         fprintf(infof, "Cylinder  Head%11s  ", temp_str);
         for (x = 0; x < sectors_last; x++)
            fprintf(infof, "%03d ", secbase_last + x);
         fprintf(infof, "\n");
        } 
     i += 3;
     fprintf(infof, "%4d       %d               ", info.buf[i], info.buf[i+1]);
     i += 2;
     for (x = 0; x < sectors_last; x++)
        {
         switch (info.buf[i])
            {
             case    0 : // sector is marked as good
                         fprintf(infof, " .  ");
                         i++;    
                         break;
             case 1000 : // sector is marked as bad
                         fprintf(infof, "ERR ");
                         i++;   
                         break;
             case 1001 : // sector was not read (aborted)
                         fprintf(infof, " X  ");
                         i++;
                         break;
               default : // sector is good but required some retries
                         retries = info.buf[i];
                         if (retries > 255)
                            retries = 255;
                         fprintf(infof, "%3d ", retries);
                         i++;    
            } 
        }
     fprintf(infof, "\n");                
    }

 // append or close the error log file if one was created
 if (errorf)
    {
     if (disk.append_error)
        {
         rewind(errorf);
         fprintf(infof, "\n");
         while ((res = fgets(temp_str, sizeof(temp_str)-1, errorf)))
            fprintf(infof, "%s", temp_str);        
         fclose(errorf);
         errorf = NULL;
         remove(error_file);
        }
     else
        fclose(errorf);        
    }  

 fclose(infof);
 infof = NULL;
}

//==============================================================================
// Place a sector read result into the info file buffer.
//
// The status of sector reads is placed in the table in the physical sector
// locations.  All sectors for each new track are filled with -1 to indicate
// a sector that has not been read and a -1 end marker inserted for the next
// track to indicate end of info.
//
//   pass: dsk_pcyl_t cyl		cylinder number
//         dsk_phead_t head		physical side of disk
//         int lsect                    logical sector number
// return: void
//==============================================================================
static void info_file_entry (dsk_pcyl_t cyl, dsk_phead_t head, int lsect)
{
 int i;
 
 if ((info.pos_trk + 10) > INFO_SIZE)
    return;

 if ((cyl != info.cyl_last) || (head != info.head_last))
    {
     info.buf[info.pos_trk++] = dg.dg_secbase;
     info.buf[info.pos_trk++] = dg.dg_sectors;
     info.buf[info.pos_trk++] = dg.dg_secsize;
     info.buf[info.pos_trk++] = cyl;
     info.buf[info.pos_trk++] = head;

     // sectors for this track will be inserted here
     info.pos_sec = info.pos_trk;

     // the next track index
     info.pos_trk += dg.dg_sectors;
     
     // set all sectors in this track with 'not read' status values
     for (i = 0; i < dg.dg_sectors; i++)
        info.buf[info.pos_sec + 1] = -2;
        
     // mark the next track as not used (end marker)
     info.buf[info.pos_trk] = -1;
     
     info.cyl_last = cyl;
     info.head_last = head;
    }

 // offset index for the physical sector to be inserted
 i = info.pos_sec + skew_table[lsect];

 // set the status of the last sector read 
 if (sect_retry_count == -1) // sector read error?
    info.buf[i] = 1000;  
 else
     if (sect_retry_count == -2) // unread sector?
        info.buf[i] = 1001;
     else  // else set the retry count
        info.buf[i] = sect_retry_count;

#if 0
 printf(" - info.pos_trk=%d, info.pos_sec=%d i=%d\n",
        info.pos_trk, info.pos_sec, i);
#endif
}

//==============================================================================
// Close all open libdsk files.
//
//   pass: void
// return: void
//==============================================================================
void close_files (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;

 if (idrive)
    {
     if (dsk_close(&idrive) != DSK_ERR_OK)
        printf(APPNAME": dsk_close(&idrive) - %s\n", dsk_strerror(dsk_err));
     idrive = NULL;
    } 

 if (odrive)
    {
     if (dsk_close(&odrive) != DSK_ERR_OK)
        printf(APPNAME": dsk_close(&odrive) - %s\n", dsk_strerror(dsk_err));
     odrive = NULL;
    } 
}

//==============================================================================
// Set the data rate and stepping for input and output floppy drives
// supported by the itype and otype drivers.
//
// The auto rate and auto double stepping sections of this function is
// really only intended for 40T 5.25" disks in a 5.25' HD PC drive.
//
// This function will also force a data rate (--datarate) and/or stepping
// rates (--idstep, --odstep) if the options are specified and will apply to
// ANY drive type.
//
// Note: If a PC has an 80T DD drive fitted (very unusual) then a disk
// definition for it using RATE_SD should be used, the autorate won't act on
// it here because it should be seen as non 'h' drive type unless the PC's
// BIOS has been set incorrectly.  Such drives should probably be set to
// 5.25" 360kb if no setting for 720k is provided.
//
//   pass: void
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
static dsk_err_t set_rate_step (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;

 // auto set the data rate and stepping for the input drive
 if (idrive && disk.autorate)
    {
     // if this is a (80T 5.25" HD or 77T 8") drive and less than 77
     // cylinders (70) then assume it to be a DD 40T 400K disk
     if (disk.idrive_type == 'h' && disk.idstep != 0 && dg.dg_cylinders < 70)
        {
         // for now input/output rate use same value
         xdg.dg_idatarate = RATE_DD;
         xdg.dg_odatarate = RATE_DD;
         dsk_err = dsk_set_option(idrive, "DOUBLESTEP", 1);
         disk.idstep_used = 1;
        }
    }

 // auto set the data rate and stepping for the output drive
 if (odrive && disk.autorate && dsk_err == DSK_ERR_OK)
    {
     // if this is a (80T 5.25" HD or 77T 8") drive and less than 77
     // cylinders (70) then assume it to be a DD 40T 400K disk
     if (disk.odrive_type == 'h' && disk.odstep != 0 && dg.dg_cylinders < 70)
        {
         // for now input/output rate use same value
         xdg.dg_idatarate = RATE_DD;
         xdg.dg_odatarate = RATE_DD;
         dsk_err = dsk_set_option(odrive, "DOUBLESTEP", 1);
         disk.odstep_used = 1;
        }
    }

 // override the data rate if a value was provided
 if (dg_opts.datarate != -1)
    {
     xdg.dg_idatarate = dg_opts.datarate;
     xdg.dg_odatarate = dg_opts.datarate;
    }

 // override the double stepping for input if a value was provided
 if ((idrive) && (dsk_err == DSK_ERR_OK) && (disk.idstep == 1))
    {
     dsk_err = dsk_set_option(idrive, "DOUBLESTEP", 1);
     disk.idstep_used = 1;
    }

 // override the double stepping for output if a value was provided
 if ((odrive) && (dsk_err == DSK_ERR_OK) && (disk.odstep == 1))
    {
     dsk_err = dsk_set_option(odrive, "DOUBLESTEP", 1);
     disk.odstep_used = 1;
    } 

 return DSK_ERR_OK;
}

//==============================================================================
// Convert macro file names and types to the format expected by the host
// allowing the ubeedisk.ini file to be portable between host platforms.
//
// Converts driver type 'fd' to the host floppy driver 'ntwdm' on Windows and
// 'floppy' on Unices.
// 
// Converts drives 'A:' - 'Z:' on Unices to devices /dev/fdx
//
// Set LibDsk driver types based on file names if the user has not specified
// the type.
//
//   pass: char *xfile
//         char *xtype
// return: void
//==============================================================================
static void set_xtype_xfile (char *xfile, char *xtype)
{
 char *p;
 int i;

 ext_type_t ext_type[] =
 {
  {".dsk",      "dsk"},
  {".edsk",    "edsk"},
  {".qm",        "qm"},
  {".tel",     "tele"},
  {".cfi",      "cfi"},
  {".nw",  "nanowasp"},
  {".raw",      "raw"},
  {".ss40",     "raw"},
  {".ss80",     "raw"},
  {".ds40",     "raw"},
  {".ds80",     "raw"},
  {".ds82",     "raw"},
  {".ds84",     "raw"},
  {".ds8x",     "raw"},
  {"",             ""}
 };

 // driver 'fd' may be used to represent the host floppy driver
 if (strcmp(xtype, "fd") == 0)
#ifdef WIN32
    strcpy(xtype, "ntwdm");
#else
    strcpy(xtype, "floppy");
#endif

 // drives 'A:' - 'Z:' may be used on Unices to represent /dev/fdx
#ifndef WIN32
 if ((strstr(xfile, ":") == xfile + 1) && (strlen(xfile) == 2) &&
 (toupper(xfile[0]) >= 'A') && (toupper(xfile[0]) <= 'Z'))
    sprintf(xfile, "/dev/fd%d", toupper(xfile[0]) - 'A');
#endif

 // do nothing if the type has been specified or no file name
 if ((xtype[0]) || (! xfile[0]))
    return;

 // determine the type required from the file name    
#ifdef WIN32
 if ((strstr(xfile, ":") == xfile + 1) && (strlen(xfile) == 2) &&
 (toupper(xfile[0]) >= 'A') && (toupper(xfile[0]) <= 'B'))
    strcpy(xtype, "ntwdm");
#else
 if (strstr(xfile, "/dev/fd") == xfile)
    strcpy(xtype, "floppy");
#endif
 else
    {
     for (i = 0; ext_type[i].ext[0]; i++)
        {
         if ((p = strstr(xfile, ext_type[i].ext)) &&
            (*(p + strlen(ext_type[i].ext)) == 0))
            {
             strcpy(xtype, ext_type[i].type);
             break;
            }
        }
    }
// printf("set_xtype_xfile: xfile=%s  xtype=%s\n", xfile, xtype);
}

//==============================================================================
// Override values
//
//   pass: void
// return: dsk_err_t			DSK_ERR_OK if successful
//==============================================================================
dsk_err_t override_values (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;

 // set the disk sidedness format
 if (dg_opts.sidedness != -1)
    dg.dg_sidedness = dg_opts.sidedness;

// set the number of cylinders
 if (dg_opts.cylinders != -1)
    dg.dg_cylinders = dg_opts.cylinders;

 // set the mumber of heads
 if (dg_opts.heads != -1)
    dg.dg_heads = dg_opts.heads;

 // set the number of sectors per track
 if (dg_opts.sectors != -1)
    dg.dg_sectors = dg_opts.sectors;

 // set the sector base 
 if (dg_opts.secbase != -1)
    {
     dg.dg_secbase = dg_opts.secbase;
     xdg.dg_secbase1s = dg_opts.secbase;
    } 

 // set the bytes per sector 
 if (dg_opts.secsize)
    dg.dg_secsize = dg_opts.secsize;

 // set the read/write gap
 if (dg_opts.rwgap != -1)
    dg.dg_rwgap = dg_opts.rwgap;

 // set the format gap
 if (dg_opts.fmtgap != -1)
    dg.dg_fmtgap = dg_opts.fmtgap;

 // set the FM mode
 if (dg_opts.fm != -1)
    dg.dg_fm = dg_opts.fm;

 // set no multi read/write
 if (dg_opts.nomulti != -1)
    dg.dg_nomulti = dg_opts.nomulti;

 // set no skip of deleted data
 if (dg_opts.noskip != -1)
    dg.dg_noskip = dg_opts.noskip;

 // set side1as0
 if (dg_opts.side1as0 != -1)
    xdg.dg_side1as0 = dg_opts.side1as0;

 // set secbase2s
 if (dg_opts.secbase2s != -1)
    xdg.dg_secbase2s = dg_opts.secbase2s;

 // set secbase2c
 if (dg_opts.secbase2c != -1)
    xdg.dg_secbase2c = dg_opts.secbase2c;

 // sector skew value
 if (dg_opts.skew_val != -1)
    xdg.dg_skew_val = dg_opts.skew_val;

 // offset to apply to sector base
 if (dg_opts.skew_ofs != -1)
    xdg.dg_skew_ofs = dg_opts.skew_ofs;

 // set special flags
 if (dg_opts.special != -1)
    xdg.dg_special = dg_opts.special;

 // set format name
 if (dg_opts.format_name[0] != 0)
    {
     strcpy(xdg.dg_format_name, dg_opts.format_name);
     strcpy(xdg.dg_format_desc, dg_opts.format_desc);
    } 

 // set sides for input
 if ((idrive) && (disk.iside != -1))
    dsk_err = dsk_set_option(idrive, "HEAD", disk.iside);

 // set sides for output
 if ((odrive) && (disk.oside != -1) && (dsk_err == DSK_ERR_OK))
    dsk_err = dsk_set_option(odrive, "HEAD", disk.oside);

 // set the data rate and stepping for the input/output drives
 if (dsk_err == DSK_ERR_OK)
    dsk_err = set_rate_step();

 // not many retries at the first level should be needed if the 'remote'
 // driver is selected as native hardware should be more reliable and having
 // a high value simply slows down the process too much if there is a bad
 // sector, when using the LibDsk driver we tend to see a lot of soft errors
 // so the value needs to be higher.
 if (disk.retries_l1 == -1)
    {
     if (strcmp("remote", disk.itype) == 0)
        disk.retries_l1 = 1;
     else
        {
         if (disk.disk == UBEEDISK_SCAN || disk.disk == UBEEDISK_SPEED)
            disk.retries_l1 = 5;
         else   
            disk.retries_l1 = 30;
        }
    }

 // set ignore all errors when in Interactive mode.
 if (disk.ignore_errors == 1)
    skip_all_errors = 1;
 else   
    skip_all_errors = 0;
    
 // this will be the first read of a sector if copying a disk
 disk.first_read = 1;

 // make a copy of the current geometry values
 memcpy(&cdg, &dg, sizeof(cdg));

 if (dsk_err != DSK_ERR_OK)
    {
     printf(APPNAME": override_values() - %s\n", dsk_strerror(dsk_err));
     disk.error = 1;
     return dsk_err;
    }

 return DSK_ERR_OK;
}

//==============================================================================
// Determine what interface support there is for an already open input or
// output drive.  This is needed by remote types FDS.COM and Floppyio to
// determine if full support for dsk_psecid(), etc is implemented.  It may
// be used for other things in the future.
//
// FDS.COM does not currently fully implement the dsk_psecid() function and
// the head value returned is what is passed to it.  Floppyio does fully
// implement the function.
//
//   pass: DSK_PDRIVER xdrive           input or output drive
//         char *xtype                  input or output type
//         sup_t *xsup                  input or output support structure
// return: void
//==============================================================================
static void interface_support (DSK_PDRIVER xdrive, char *xtype, sup_t *xsup)
{
 char *p[1];
 int i;

 // clear all the support values
 memset(xsup, 0, sizeof(xsup));

 // if not the 'remote' type
 if (strcmp("remote", xtype))
    {
     // if not the 'raw' type
     if (strcmp("raw", xtype))
        {
         xsup->psecid = 1;
         xsup->ptrackids = 1;
         xsup->xread = 1;
         xsup->xwrite = 1;
        }

     xsup->pformat = 1;
     strcpy(xsup->version, "n/a");
     strcpy(xsup->remote_interface, "n/a");
     return;
    }

 // FDS.COM and Floppyio partially support xread and xwrite 
 xsup->xread = 1;
 xsup->xwrite = 1;

 for (i = 0; ;i++)
    {
     dsk_option_enum(xdrive, i, p);
     if (*p && strstr(*p, "Interface:") == *p)
        {
         strcpy(xsup->remote_interface, *p + 11);
         if (! strcmp(xsup->remote_interface, "Floppyio"))
            {
             xsup->psecid = 1;
             xsup->ptrackids = 1;
             xsup->pformat = 1;

             // if required the next enumeration should hold the version
             i++;
             dsk_option_enum(xdrive, i, p);
             if (*p)
                strcpy(xsup->version, *p + 9);
            }
         return;
        }
     else
        if (! *p)
           return;
    }
}

//==============================================================================
// Open input disk/image drive.
//
// Opens input drive to a default format or another if specified in
// disk.format.
//
//   pass: void
// return: int				0 if successful else -1
//==============================================================================
static int open_input_drive (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;

 char *incomp;
 
 // open the input drive
 if (disk.incomp[0])
    incomp = disk.incomp;
 else
    incomp = NULL;
 
 if (! disk.itype[0])
    {
     printf(APPNAME": open_input_drive(), use the --itype option to"
            " specify the input type\n");
     disk.error = 1;
     return -1;
    }

 reset_drive(disk.itype, disk.ifile);

 dsk_err = dsk_open(&idrive, disk.ifile, disk.itype, incomp);

 if (dsk_err != DSK_ERR_OK)
    {
     printf(APPNAME": open_input_drive() - %s\n", dsk_strerror(dsk_err));
     disk.error = 1;
     return -1;
    }

 disk.idrive_type = get_drive_type(disk.itype, disk.ifile, idrive);
 if (disk.verbose > 1)
    printf("Input drive type: %c (char) %d (dec)\n",
           disk.idrive_type, disk.idrive_type);

 interface_support(idrive, disk.itype, &input_sup);
 if (disk.verbose > 1)
    {
     printf("Input interface name: %s\n", input_sup.remote_interface);
     printf("Input interface version: %s\n", input_sup.version);
    }

 return 0;
}

//==============================================================================
// Create output disk/image drive.
//
// Creates an output disk using the same geometry as the input drive
//
//   pass: void
// return: int				0 if successful else -1
//==============================================================================
static int create_output_drive (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;

 char *outcomp;

 if (disk.outcomp[0])
    outcomp = disk.outcomp;
 else
    outcomp = NULL;
 
 if (! disk.otype[0])
    {
     printf(APPNAME": create_output_drive(), use the --otype option to"
            " specify the output type\n");
     disk.error = 1;
     return -1;
    }

 reset_drive(disk.otype, ofile_name);
 
 dsk_err = dsk_creat(&odrive, ofile_name, disk.otype, outcomp);

 if (dsk_err != DSK_ERR_OK)
    {
     printf(APPNAME": create_output_drive() - %s\n", dsk_strerror(dsk_err));
     disk.error = 1;
     return -1;
    }

 disk.odrive_type = get_drive_type(disk.otype, ofile_name, odrive);
 if (disk.verbose > 1)
    printf("Output drive type: %c (char) %d (dec)\n",
           disk.odrive_type, disk.odrive_type);

 // retry value of 1 only writes once
 dsk_set_retry(odrive, 10);

 interface_support(odrive, disk.otype, &output_sup);
 if (disk.verbose > 1)
    {
     printf("Output interface name: %s\n", output_sup.remote_interface);
     printf("Output interface version: %s\n", output_sup.version);
    }

 return 0;
}

//==============================================================================
// Open input and output disk/image drives.
//
// Opens input drive to a default format or another if specified in
// disk.format.
//
//   pass: void
// return: int				0 if successful else -1
//==============================================================================
static int open_drives (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;

 dsk_format_t fmt; 
 dsk_cchar_t fdesc;
 dsk_cchar_t fname;
 
 int res = -1;
 disk.error = 0;

 // set default common geometry values
 format_set_geometry(&microbee_disk_format[fmt_default]);

 // open an input file if one is specified
 if (disk.ifile[0])
    {
     dsk_err = open_input_drive();
     if (dsk_err != DSK_ERR_OK)
        {
         disk.error = 1;
         return -1;
        }
    }     

 // if scanning then no need for any special format
 if (disk.disk == UBEEDISK_SCAN ||
     disk.disk == UBEEDISK_SPEED ||
     disk.disk == UBEEDISK_CLEAN)
    return 0;   

 // open the output file if one is specified
 if (ofile_name[0])
    {
     dsk_err = create_output_drive();
     if (dsk_err != DSK_ERR_OK)
        {
         disk.error = 1;
         return -1;
        }
    }     

 if ((! disk.detect) && (! disk.format[0]))
    {
     printf(APPNAME": no disk format! use --detect or --format options.\n");
     disk.error = 1;
     return -1;
    } 

 // check if a format is an internal type provided by ubeedisk
 if (disk.format[0])
    res = format_set(disk.format);

 // try and detect the disk format
 if ((disk.detect) && (! disk.format[0]))
    res = format_detect_disk();

 if ((res == -1) && (disk.format[0])) // if no format or detection found
    {
     // find the requested format and use that
     int found = 0;
     fmt = FMT_180K;
     while ((dg_stdformat(NULL, fmt, &fname, NULL) == DSK_ERR_OK) && (! found))
        {
         if (! strcmp(disk.format, fname))
            found = 1;
         else
            ++fmt;
        }

     if (! found)
        {   
         printf(APPNAME": format name '%s' not recognised.\n", disk.format);
         disk.error = 1;
         return -1;
        }
       
     // initialise disk geometry                                                       
     dsk_err = dg_stdformat(&dg, fmt, NULL, &fdesc);
     if (dsk_err == DSK_ERR_OK)
        {
         res = 0;
         if (disk.verbose)
            printf("\nLibDsk/libdskrc format selected:\n%s: %s\n",
                   fname, fdesc);
        } 
     else   
        {
         printf(APPNAME": format name '%s' not recognised.\n", fdesc);
         disk.error = 1;
         return -1;
        }
    } 

 if (res == -1)
    {
     disk.error = 1;
     return -1;
    } 

 return 0;
}

//==============================================================================
// Reset input drive and settings for PC Floppy disks.
//
// LibDsk seems to get confused and won't recover if no disk is in the drive
// when it tries to read it (Input/output error).  Closing and reopening the
// input disk corrects for this on a PC Floppy.
//
// This function may also be called to overcome other problems when the
// drive is not ready.
//
// Don't do this if using any other input type (especially Remote) as it can
// cause a time-out and it should not be needed.
//
//   pass: void
// return: void
//==============================================================================
void reset_input_drive_and_settings (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;

#ifdef WIN32
 if (strcmp("ntwdm", disk.itype) != 0)
#else
 if (strcmp("floppy", disk.itype) != 0)
#endif
    return;
    
 if (dsk_close(&idrive) != DSK_ERR_OK)
    {
     printf(APPNAME": reset_input_drive_and_settings() - %s\n",
     dsk_strerror(dsk_err));
     exit(EXIT_FAILURE);
    }
    
 idrive = NULL;

 if (open_input_drive() != 0)
    exit(EXIT_FAILURE);

 override_values();
}

//==============================================================================
// Enter a disk description.
//
// Up to DESC_LINES lines may be entered.  Terminate by entering a '.' on a
// line by itself.  The last line always contains '.\n' by iteself.
//
//   pass: void
// return: void
//==============================================================================
static void enter_disk_description (void)
{
 int i;

 // We will be prompted for a disk description if the following rule is true:
 //
 // If the 'info' file is enabled (default=enabled) AND
 // options --entdesc=on (default=on) AND
 // options --diskdesc OR --diskdescf have NOT been specified AND
 // The output type is NOT in the 'no_desc_otypes[]' list.

 if ((! disk.info_file) || (! disk.enter_desc) ||
 (string_search(no_desc_otypes, disk.otype) != -1))
    return;
 
 description[0][0] = 0; 
 
 printf("\n"
        "Enter a disk description,  a '.' on a line by itself ends input."
        "  Maximum %d\n"
        "lines allowed. (Cut and paste from a text editor to make this easier)"
        "\n\n", DESC_LINES);

 for (i = 0; i < DESC_LINES; i++)
    {
     if (fgets(description[i], DESC_CHARS, stdin)) {} // ignore result 
     if (strcmp(".\n", description[i]) == 0)
        break;
    }
 description[i][0] = 0;    
}

//==============================================================================
// Get a disk description from a text file.
//
// Prompting for a disk description will be disabled.
//
// This works the same way as for the enter_disk_description() function
// except takes the input from a file specified with an option.
//
//   pass: char *s                      disk description file name
// return: int                          -1 if error, else 0
//==============================================================================
int file_disk_description (char *s)
{
 FILE *fp;
 int i;

 disk.enter_desc = 0;  // disable prompting for a disk description.

 description[0][0] = 0; 

 fp = fopen(s, "r");
 if (! fp)
    return -1;
    
 for (i = 0; i < DESC_LINES; i++)
    {
     if (fgets(description[i], DESC_CHARS, fp)) {} // ignore result 
     if (strcmp(".\n", description[i]) == 0)
        break;
    }
 description[i][0] = 0;

 fclose(fp);
 return 0;
}

//==============================================================================
// Process a disk description line of text passed from an option.
//
// This function is called repeatedly for each line of text to be added to
// the disk description.  Prompting for a disk description will be disabled.
//
// Up to DESC_LINES lines may be entered and each line may contain up to
// DESC_CHARS characters.  Any excess characters will be lost so it will be
// up to the caller to make sure lines are passed one at a time and not as
// one long string.
//
// The last line must contain '.\n' by iteself and is inserted after each
// call to this function and overwritten on the next call.
//
//   pass: char *s                      disk description string
// return: void
//==============================================================================
void process_disk_description (char *s)
{
 disk.enter_desc = 0;  // disable prompting for a disk description.

 if (diskdesc_line == DESC_LINES)
    return;
    
 strncpy(&description[diskdesc_line][0], s, DESC_CHARS);
 strcat(&description[diskdesc_line++][0], "\n");

 // terminate the lines of input with a '.\n' on a line by itself
 strcpy(&description[diskdesc_line][0], ".\n");
}

//==============================================================================
// Set start and finish cylinder values.
//
//   pass: int *cyl_start
//         int *cyl_finish
// return: void
//==============================================================================
static void set_start_finish (int *start, int *finish)
{
 // starting cylinder
 if (disk.start < 0)
    *start = 0;
 else
    *start = disk.start;

 // finishing cylinder
 if (disk.finish < 0)
    *finish = dg.dg_cylinders - 1;
 else   
    *finish = disk.finish;
}

//==============================================================================
// Set special disk values.
//
//   pass: dsk_pcyl_t cyl		cylinder number
//         dsk_phead_t head		physical side of disk
//         int secsize                  sector size
// return: void
//==============================================================================
static void set_special_disk (dsk_pcyl_t cyl, dsk_phead_t head, int secsize)
{
 // Honeysoft format ?
 if (xdg.dg_special & FLAGS_HSOFT)
    {
     if (secsize != -1)
        dg.dg_secsize = secsize;
     else
        {
         if (cyl >= 5)
            dg.dg_secsize = 1024;
         else
            dg.dg_secsize = 512;
        }
     if (dg.dg_secsize == 1024)
        dg.dg_sectors = 5;
     else
        dg.dg_sectors = 10;
     // we should increase the gap values for 1024 byte sectors at some stage.   
    }
    
 // C5H0S0 disk format ?
 if (xdg.dg_special & FLAGS_C5H0S0)
    {
     if (cyl == 5 && head == 0)
        dg.dg_secbase = 0;
     else
        dg.dg_secbase = 1;
    }

 // 8" 60 sector disk format ?
 if (xdg.dg_special & FLAGS_60)
    {
     if (head == 0)
        dg.dg_secbase = 1;
     else
        dg.dg_secbase = 31;
    }
}

//==============================================================================
// Disk clean.
//
// This may be used to clean the drive heads using a special cleaning disk.
//
// As there is no low level control disk cleaning will simply try to read a
// sector ID at given a given cylinder and side.  The retry value will be
// set as low as possible.  After the call times-out the process will be
// repeated for other tracks.
//
// This function may be called as a command line process or while a disk is
// being copied.
//
//   pass: int method                   pass 'c' if command line
// return: void
//==============================================================================
static void disk_clean (int method)
{
 char user[100];

 DSK_FORMAT result; 

 dsk_pcyl_t cyl;
 int iterations = 0;
 int done = 0;
 
 // check that the input drive is a floppy or remote type
 if (string_search(no_info_file, disk.itype) == -1)
    {
     printf("\n"APPNAME": input device must be a floppy drive for cleaning!\n");
     return;
    } 

 if (method == 'c')
    {
     // set input type based on input name
     set_xtype_xfile(disk.ifile, disk.itype);

     if (open_drives() == -1)
        return;

     if (override_values() != DSK_ERR_OK)
        return;

     if (disk.verbose)
        printf("\n");
    }
     
 // if the input driver does not support the dsk_psecid() function then
 // exit immidiately as there is no point in continuing.
 if (! input_sup.psecid)
    {
     printf(APPNAME": input device does not support the"
            " dsk_psecid() function.\n");
     return;
    } 

 // set 0 retries for the input drive (we want it to give up ASAP)
 dsk_set_retry(idrive, 0);

 // we must set some valid values if this is the command line method
 if (method == 'c')
    {
     dg.dg_fm = 0;
     dg.dg_heads = 255;
     dg.dg_noskip = 1;
     dg.dg_datarate = RATE_SD;
    }
 else
    {
     printf("\nInsert the cleaning disk then press ENTER to start the"
            " cleaning process.\n");
     if (fgets(user, sizeof(user)-2, stdin))
        {} // ignore result
    }   

 printf("Press any key to exit (exits after 5 passes)...\n\n");

 while (iterations++ != 5)
    {
     printf("Cleaning pass: %d/5\n", iterations); 

     // step in 5 cylinders at a time
     for (cyl = 0; cyl < 40;)
        {
         if (disk.verbose > 1)
            printf("Moving head to cylinder %2d.\n", cyl); 
         done = get_key() != -1;
         if (done)
            break;
         dsk_psecid(idrive, &dg,  cyl, 0, &result);
         sleep_ms(100);
         cyl += 5;
        }

     if (done)
        break;

     // step back 5 cylinders at a time
     for (cyl = 30; cyl > 0;)
        {
         if (disk.verbose > 1)
            printf("Moving head to cylinder %2d.\n", cyl); 
         done = get_key() != -1;
         if (done)
            break;
         dsk_psecid(idrive, &dg,  cyl, 0, &result);
         sleep_ms(100);
         cyl -= 5;
        }

     if (done)
        break;
    }

 if (method == 'c')
    {
     printf("\nRemove the cleaning disk from the drive.\n"); 
     close_files();
     return;
    }

 printf("\nRemove the cleaning disk and REPLACE disk.\n");

 // flush any pending keys;
 while (get_key() != -1)
    ;

 // must close and reopen the input drive and restore settings.
 reset_input_drive_and_settings();

 // need to put this value back  
 dsk_set_retry(idrive, disk.retries_l1);
}

//==============================================================================
// Disk scan head options.
//
// --scanhead=s[,f]
//
// Set the head range to be scanned when using the --disk=scan option, 's'
// is the head to start with, 'f' is an optional finish head otherwise
// finish will be the same as 's'.  Default values are are 0,1.
//
//   pass: char *parms                  command parameters
// return: int                          0 if no error, else -1
//==============================================================================
int scan_option (char *parms)
{
 char s[100];
 char *c;
 int start;
 int finish;
 
 // get the head start number 's'
 c = get_next_parameter(parms, ',', s, &start, sizeof(s)-1);
 if (start < 0 || start > 255)
    return -1;
 
 disk.head_s = start;
 disk.head_f = start;

 // if no parameter given for the optional 'f' head then we are done
 if (! c)
    return 0;

 // get the head finish number 'f'
 c = get_next_parameter(c, ',', s, &finish, sizeof(s)-1);
 if (finish < 0 || finish > 255)
    return -1;

 // check that the finish head is higher or same as the start head
 if (finish < start)
    return -1;
    
 disk.head_f = finish;    

 // can't be any more sub-parameters
 if (c)
    return -1;

 return 0;
}

//==============================================================================
// Disk scan.
//
// Scans a disk with auto detection for data rate and encoding method.
//
// The output reported is from the sector ID on the disk and these values
// may differ to the physical Cylinder and Head values.
//
//   pass: void
// return: void
//==============================================================================
static void disk_scan (void)
{
 // data rates arranged in order most likely to be found first
 dsk_rate_t datarates[4] = {RATE_SD, RATE_DD, RATE_HD, RATE_ED};

 // these must be kept in this order
 char datarates_str[4][5] = {" 500", " 250", " 300", "1000"};
 char encmode_str[2][4] = {"mfm", " fm"};

 dsk_err_t dsk_err = DSK_ERR_OK;

 dsk_psect_t count;
 DSK_FORMAT *result;
 DSK_FORMAT sector_id;

 dsk_pcyl_t cyl;
 dsk_phead_t head;
 int i;
 int dr;

 // set input type based on input name
 set_xtype_xfile(disk.ifile, disk.itype); 

 // check we have an input file name
 if (! disk.ifile[0])
    {
     printf(APPNAME": use '--if' option to specify an input file.\n");
     disk.error = 1;
    }

 if (open_drives() == -1)
    return;

 // if the input driver does not support the dsk_ptrackids() function then
 // exit as there is no point in continuing.
 if (! input_sup.ptrackids)
    {
     printf(APPNAME": input device does not support the"
            " dsk_ptrackids() function.\n");
     return;
    } 

 if (override_values() != DSK_ERR_OK)
    return;

 if (disk.verbose)
    printf("\n");

 // set starting and finishing cylinder values
 set_start_finish(&cyl_start, &cyl_finish);

 // set retries for the input drive
 dsk_set_retry(idrive, disk.retries_l1);
 
 // if supported by the driver disable the AUTORATE detection
 dsk_set_option(idrive, "AUTORATE", 0);

 // The first time the head may need to be homed and this can cause a non
 // recoverable error so do a dummy read of T0, H0, if error close, delay and
 // open again.
 dg.dg_fm = 0;
 dg.dg_heads = 255;
 dg.dg_noskip = 1;
 dg.dg_datarate = datarates[0];

 // some drives need this or this disk head won't get positioned and we end
 // up with no data error! (must be before reset_input_drive_and_settings())
 dsk_err = dsk_psecid(idrive, &dg, 0, 0, &sector_id);

 // must close and reopen the input drive and restore settings.
 reset_input_drive_and_settings();
 dsk_set_option(idrive, "AUTORATE", 0);

 dsk_err = dsk_psecid(idrive, &dg, 0, 0, &sector_id);

 // must close and reopen the input drive and restore settings if error.
 if (dsk_err != DSK_ERR_OK)
    {
     sleep_ms(2000);
     reset_input_drive_and_settings();
    }

 for (cyl = cyl_start; cyl <= cyl_finish; cyl++)
    {
     for (head = disk.head_s; head <= disk.head_f; head++)
        {
         // check all data rates with MFM (0) mode first then FM (1)
         for (dg.dg_fm = 0; dg.dg_fm < 2; ++dg.dg_fm)
            {
             // find the data rate starting with the most likely first
             for (dr = 0; dr < 4; ++dr)
                {
                 dg.dg_datarate = datarates[dr];
                 dsk_err = dsk_psecid(idrive, &dg, cyl, head, &sector_id);
                 if (! dsk_err)
                    break;
                }	
             if (! dsk_err)
                break;
            }

         if (dsk_err)
            printf("Nothing detected: Cylinder: %3d  Head: %2d"
                   "  Result: %s\n\n", cyl, head, dsk_strerror(dsk_err));
         else
            {
             dsk_err = dsk_ptrackids(idrive, &dg, cyl, head, &count, &result);

             if (dsk_err == DSK_ERR_OK)
                {
                 printf("Cylinder: %3d  Head: %2d  Sectors: %2d"
                        "  Data rate kHz: %s  Encoding: %s\n",
                        cyl, head, count, datarates_str[dg.dg_datarate],
                        encmode_str[dg.dg_fm]);
                 printf("------------------------------------"
                        "------------------------------------\n");
         
                 for (i = 0; i < count; i++)
                    {
                     sector_id = result[i];
                     printf("Track: %3d  Side: %2d  Sector: %3d"
                            "  Bytes/sector: %4d\n",
                            sector_id.fmt_cylinder,
                            sector_id.fmt_head,
                            sector_id.fmt_sector,
                            sector_id.fmt_secsize);
                    }
                 dsk_free(result);   
                 printf("\n");
                }
             else
                {
                 // depending on the error terminate the scan process
                 // printf(APPNAME": error reading track IDs: %s\n",
                 // dsk_strerror(dsk_err));
                 // return;
                }
            }

         // allow the operator to pause, resume, abort, and clean
         if (pause_menu(1) == DSK_ERR_ABORT)
            {
             close_files();
             return;
            }
        }
    }
 close_files();
}

//==============================================================================
// Format a disk/image(s).
//
// A range of cylinders can be formatted.
//
//   pass: void
// return: void
//==============================================================================
static void disk_format (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;
 dsk_pcyl_t cyl;
 dsk_phead_t head;

 char user[10];

 // must make sure there is no input file or strange things happen!
 disk.ifile[0] = 0;

 // set output type based on output name
 set_xtype_xfile(disk.ofile, disk.otype); 

 // check we have an output file name.
 if (! disk.ofile[0])
    {
     printf(APPNAME": use --of option to specify an output file.\n");
     return;
    }
 
 if (disk.noformat)
    {
     printf("\n"APPNAME": '--noformat' option is enabled, no format!\n");
     return;
    } 

 strcpy(ofile_name, disk.ofile);

 if (disk.verbose)
    printf("\nFormatting disk/image '%s'\n", ofile_name);

 // prompt the user to confirm format if no --force option specified
 if (! disk.force)
    {
     user[0] = 0;
     printf("\nWARNING: To format type 'format', anything else aborts: ");
     if (fgets(user, sizeof(user), stdin)) {} // ignore result 
     if (strcmp("format\n", user) != 0)
        {
         printf("Format aborted.\n");
         return;
        }
    }

 if (open_drives() == -1)
    return;

 // if the output driver does not support the dsk_pformat() function then
 // exit as there is no point in continuing.
 if (! output_sup.pformat)
    {
     printf(APPNAME": output device does not support the"
            " dsk_pformat() function.\n");
     return;
    } 

 if (override_values() != DSK_ERR_OK)
    return;

 if (disk.verbose)
    printf("\n");

 // set starting and finishing cylinder values
 set_start_finish(&cyl_start, &cyl_finish);

 dsk_set_retry(odrive, 10);

 // set data rate for output
 dg.dg_datarate = xdg.dg_odatarate;
 
 // format the requested cylinders
 for (cyl = cyl_start; cyl <= cyl_finish; cyl++)
    {
     for (head = 0; head < dg.dg_heads; head++)
        {
         if ((xdg.dg_secbase2c != -1) && (cyl >= xdg.dg_secbase2c))
            dg.dg_secbase = xdg.dg_secbase2s;
         else   
            dg.dg_secbase = xdg.dg_secbase1s;
            
         // handle special disk formats
         set_special_disk(cyl, head, -1);

         if (disk.verbose)         
            {
             printf("\rCylinder: %02d/%02d Head: %d/%d",
                    cyl, dg.dg_cylinders-1, head, dg.dg_heads-1);
             fflush(stdout);
            } 
            
         switch (disk.forceside)
            {
             case 0 : // off
                dsk_err = format_track(cyl, head, head & (1 ^ xdg.dg_side1as0));
                break;
             case 1 : // on
                // force side ID to be equal to the physical side
                dsk_err = format_track(cyl, head, head);
                break;
             case 2 : // 00
                dsk_err = format_track(cyl, head, 0);
                break;
             case 3 : // 01
                if (head == 0)
                   dsk_err = format_track(cyl, head, 0);
                else
                   dsk_err = format_track(cyl, head, 1);
                break;
             case 4 : // 10
                if (head == 0)
                   dsk_err = format_track(cyl, head, 1);
                else
                   dsk_err = format_track(cyl, head, 0);
                break;
             case 5 : // 11
                dsk_err = format_track(cyl, head, 1);
                break;
            }

         if (dsk_err != DSK_ERR_OK)
            return;
         
         // allow the operator to pause, resume and abort
         if (pause_menu(2) == DSK_ERR_ABORT)
            {
             printf("\n");
             return;
            }
        }
    }

 if (disk.verbose)
    printf("\n\nFormat completed.\n");
}

//==============================================================================
// Disk speed.
//
// Measure rotational speed of disk.  This is done by repeatedly reading the
// same sector and measuring the time between the first and last read.  A
// standard disk must be used that does not have the same sector on the
// track more than once.
//
//   pass: void
// return: void
//==============================================================================
static void disk_speed (void)
{
 // data rates arranged in order most likely to be found first
 dsk_rate_t datarates[4] = {RATE_SD, RATE_DD, RATE_HD, RATE_ED};

 // these must be kept in this order
 char datarates_str[4][5] = {" 500", " 250", " 300", "1000"};
 char encmode_str[2][4] = {"mfm", " fm"};

 dsk_err_t dsk_err = DSK_ERR_OK;

 dsk_psect_t count;
 DSK_FORMAT *result;
 DSK_FORMAT sector_id; 

 dsk_pcyl_t cyl;
 dsk_phead_t head;
 dsk_psect_t psect;
 int dr;

 int i;
 int iterations = 0;
 int rpm;
 int64_t ms_start;
 int64_t ms_finish;
 int ms_total;

 // set input type based on input name
 set_xtype_xfile(disk.ifile, disk.itype); 

 // check we have an input file name
 if (! disk.ifile[0])
    {
     printf(APPNAME": use '--if' option to specify an input file.\n");
     disk.error = 1;
    }

 if (open_drives() == -1)
    return;

 // if the input driver does not support the dsk_psecid() function then
 // exit as there is no point in continuing.
 if (! input_sup.psecid)
    {
     printf(APPNAME": input device does not support the"
            " dsk_psecid() function.\n");
     return;
    } 

 if (override_values() != DSK_ERR_OK)
    return;

 if (disk.verbose)
    printf("\n");

 // set starting and finishing cylinder values. Only the start value is used.
 set_start_finish(&cyl_start, &cyl_finish);
 cyl = cyl_start;

 // speed testing can be instructed what side of the disk to read with the
 // --iside option.  If set to default of -1 then uses side 0.
 if (disk.iside == -1)
    head = 0;
 else
    head = disk.iside;

 // set retries for the input drive
 dsk_set_retry(idrive, disk.retries_l1);
 
 // if supported by the driver disable the AUTORATE detection
 dsk_set_option(idrive, "AUTORATE", 0);

 // The first time the head may need to be homed and this can cause a non
 // reverable error so do a dummy read of T0, H0, if error close, delay and
 // open again.
 dg.dg_fm = 0;
 dg.dg_heads = 255;
 dg.dg_noskip = 1;
 dg.dg_datarate = datarates[0];
 dsk_err = dsk_psecid(idrive, &dg, 0, 0, &sector_id);

 // must close and reopen the input drive and restore settings if error.
 if (dsk_err != DSK_ERR_OK)
    {
     sleep_ms(2000);
     reset_input_drive_and_settings();
     dsk_set_option(idrive, "AUTORATE", 0);
    }
 
 // check all data rates with MFM (0) mode first then FM (1)
 for (dg.dg_fm = 0; dg.dg_fm < 2; ++dg.dg_fm)
    {
     // find the data rate starting with the most likely first
     for (dr = 0; dr < 4; ++dr)
        {
         dg.dg_datarate = datarates[dr];
         dsk_err = dsk_psecid(idrive, &dg, cyl, head, &sector_id);
         if (! dsk_err)
            break;
        }	
     if (! dsk_err)
        break;
    }

 if (dsk_err)
    {
     printf("Nothing detected: Cylinder: %3d  Head: %2d  Result: %s\n",
     cyl, head, dsk_strerror(dsk_err));
     printf("Note: A disk that has at minimum track 0 and side 0 formatted"
            " must be in the drive.\n");
     close_files();         
     return;
    }

 dsk_err = dsk_ptrackids(idrive, &dg, cyl, head, &count, &result);

 if (dsk_err)
    {
     close_files();         
     return;
    }

 printf("Cylinder: %3d  Head: %2d  Sectors: %2d"
        "  Data rate kHz: %s  Encoding: %s\n",
        cyl, head, count, datarates_str[dg.dg_datarate],
        encmode_str[dg.dg_fm]);
 
 psect = sector_id.fmt_sector;
 
 printf("Disk rotations for results: 5  Sector number sampled: %d\n\n", psect);

 // now that we have all the values from the unkown disk format we can read
 // the same sector over and over to find the rotational speed of the drive.
 printf("Press any key to pause (exits after 60 iterations)...\n\n");

 while (iterations++ != 60)
    {
     // allow the operator to pause, resume, abort, and clean
     if (pause_menu(1) == DSK_ERR_ABORT)
        break;

     // set the data rate back to what it was (in case of clean)        
     dg.dg_datarate = datarates[dr];

     // get the first occurrence of the sector
     for (;;)
        {
         dsk_err = dsk_psecid(idrive, &dg, cyl, head, &sector_id);
         if (psect == sector_id.fmt_sector || dsk_err != DSK_ERR_OK)
            break;
        }    
     ms_start = time_get_ms();

     // read in the same sector 5 ot 6 times depending on the drive type
     for (i = 0; i < 5; i++)
        {
         for (;;)
            {
             dsk_err = dsk_psecid(idrive, &dg, cyl, head, &sector_id);
             if (psect == sector_id.fmt_sector || dsk_err != DSK_ERR_OK)
                break;
            }
         }   

     ms_finish = time_get_ms();
     ms_total = ms_finish - ms_start;

     // add 0.5 to round the result up
     rpm = (int)(300.0 * (1000.0 / ms_total) + 0.5);

     printf("Iterations: %2d  Time (ms): %4d  RPM: %3d\n",
     iterations, ms_total, rpm);
    }

 close_files();
}

//==============================================================================
// Copy one disk/image(s).
//
// Retries (3 levels) - An explanation of how the 3 levels of retries works
// is provided here:
//
// LibDsk (level 1)
// ----------------
// A LibDsk retry level is set here to help make sure there is plenty of
// time for a sector to be read successufully at the LibDsk level.  This
// should not be set to a too low value as it causes many soft errors and
// possibly causes read problems on some disks (especially Microbee disks). 
// The value can be changed using the --retry-l1 option if required.
//
// read_sector() (level 2)
// -----------------------
// The 2nd level of retries are performed in read_sector() which can be
// changed using the --level-l2 option if required.  If the LibDsk retries
// fail then this retry level comes into play.
//
// read_sector_retry() (level 3)
// -----------------------------
// If level 2 fails to read a sector then this last level comes into play.
//
// If the Unattended mode is in use (default) then some more retries (with
// and without T0 seeks), ignore or abort operations will come into play. 
// The exact way this works depends on the count of previous errors and seek
// counts that have already occured.
//
// If interactive mode is in use then the operator has complete control and
// makes a selection when prompted for the next course of action.
//
//   pass: void
// return: void
//==============================================================================
static void copy_one_disk (void)
{
 dsk_err_t dsk_err = DSK_ERR_OK;
 dsk_err_t dsk_err_wrt = DSK_ERR_OK;
 DSK_FORMAT sector_id; 

 uint8_t *p;
 dsk_pcyl_t cyl;
 dsk_phead_t head;
 dsk_phead_t xhead;

 char trk_str[5001]; 
 int trk_str_len;

 int xsecsize;
 int lsect;
 int psect;
 int aborted = 0;

 if (open_drives() == -1)
    return;

 if (override_values() != DSK_ERR_OK)
    return;

 // initialise the info file
 init_info_file();
 
 // enter a disk description
 enter_disk_description();

 if (disk.verbose) 
    printf("\nCopying disk/image '%s' to '%s'\n", disk.ifile, ofile_name);

 sect_errors_tot = 0;
 sect_retries_tot = 0;

 auto_retry_abort = 0;
 auto_cyl_last = -1;
 auto_head_last = -1;
 auto_seeked_count = 0;

 // set starting and finishing cylinder values
 set_start_finish(&cyl_start, &cyl_finish);

 // format starting cylinders that are being skipped (LibDsk insists)
 if (cyl_start > 0)
    {
     if (disk.verbose)
        printf("\n");
     format_track_range(0, cyl_start-1);
    } 

 if (disk.verbose)
    {
     if (disk.noformat || ! output_sup.pformat)
        printf("\nCopying data:\n");
     else
        printf("\nCopying data (with format):\n");
    }

 // some drives need this or theis disk head won't get positioned and we end
 // up with no data error! (must be before reset_input_drive_and_settings())
 dsk_psecid(idrive, &dg, 0, 0, &sector_id);

 // must close and reopen the input drive and restore settings.
 reset_input_drive_and_settings();

 for (cyl = cyl_start; cyl <= cyl_finish; cyl++)
    {
     for (head = 0; head < dg.dg_heads; head++)
        {
         if ((xdg.dg_secbase2c != -1) && (cyl >= xdg.dg_secbase2c))
            dg.dg_secbase = xdg.dg_secbase2s;
         else   
            dg.dg_secbase = xdg.dg_secbase1s;

         // read first available sector ID to get the side ID
         if (! aborted)
            dsk_err = read_sector_id(cyl, head, &xhead, &xsecsize);
         else
            {
             xhead = head;
             xsecsize = dg.dg_secsize;
             dsk_err = DSK_ERR_OK;
            } 

         // handle special disk formats
         set_special_disk(cyl, head, xsecsize);
         
         // create the skew table everytime as the format may change
         create_skew_table(xdg.dg_skew_val, xdg.dg_skew_ofs, dg.dg_sectors);
         
         // format one track
         if (! disk.noformat && aborted != 2)
            {
             switch (disk.forceside)
                {
                 case 0 : // off
                    if (input_sup.psecid)
                       // follow the input sector header ID value (xhead) if a
                       // true psecid is supported
                       dsk_err = format_track(cyl, head, xhead);
                    else
                       // else the format's side1as0 value will determine
                       dsk_err = format_track(cyl, head, xhead &
                                 (1 ^ xdg.dg_side1as0));
                    break;
                 case 1 : // on
                    // force side ID to be equal to the physical side
                    dsk_err = format_track(cyl, head, head);
                    break;
                 case 2 : // 00
                    dsk_err = format_track(cyl, head, 0);
                    break;
                 case 3 : // 01
                    if (head == 0)
                       dsk_err = format_track(cyl, head, 0);
                    else
                       dsk_err = format_track(cyl, head, 1);
                    break;
                 case 4 : // 10
                    if (head == 0)
                       dsk_err = format_track(cyl, head, 1);
                    else
                       dsk_err = format_track(cyl, head, 0);
                    break;
                 case 5 : // 11
                    dsk_err = format_track(cyl, head, 1);
                    break;
                }
            }

         // read and write one complete track
         buffered_cylinder = -1;
         buffered_head = -1;
         
         trk_str_len = 0;

         if (dsk_err == DSK_ERR_OK)
            {
             // read the track
             for (lsect = 0; lsect < dg.dg_sectors; lsect++)
                {
                 // get the skewed physical sector number
                 psect = skew_table[lsect] + dg.dg_secbase;
             
                 // read sectors and write to info file
                 if (! aborted)
                    {
                     dsk_err = read_sector_retry(cyl, head, xhead, lsect);
                     if (disk.verbose > 1)
                        trk_str_len += sprintf(trk_str + trk_str_len,
                        " %2d", psect);
                    }
                 else
                    {
                     // calculate the buffer pointer location
                     p = buf + dg.dg_secsize * (psect - dg.dg_secbase);

                     // fill the sector buffer location with 0xff
                     memset(p, 0xff, dg.dg_secsize);
                     
                     // unread sector status
                     sect_errors_tot++;
                     sect_retry_count = -2;
                    }

                 // log the status of each sector read
                 info_file_entry(cyl, head, lsect);

                 // if read sector retry returned an abort repsonse
                 if (dsk_err == DSK_ERR_ABORT)
                    {
                     // set the abort method depending on the output type
                     aborted = (string_search(no_info_file, disk.otype) == -1) ? 1 : 2;
                     dsk_err = DSK_ERR_OK;
                    } 
                }

             // output the physcial sector string                
             if (disk.verbose > 1 && ! aborted)
                printf("  %s\n", trk_str);

             // write the buffered track
             if (aborted != 2)
                dsk_err_wrt = write_buffered_track(cyl, cyl, head, xhead);

             if (dsk_err_wrt != DSK_ERR_OK)
                {
                 if (disk.verbose > 1)
                    printf("\nWrite error: %s\n", dsk_strerror(dsk_err_wrt));
                }
            }
        }
    }

 if (disk.verbose)
    printf("\n");

 // format remaining cylinders that are being skipped
 if (cyl_finish < dg.dg_cylinders-1)
    {
     if (disk.verbose)
        printf("\n");
     format_track_range(cyl_finish+1, dg.dg_cylinders-1);
    }

 // output abort abort message
 if (aborted)
    {
     if (disk.unattended)
        printf("\nUnattended mode aborted this disk copy as the"
               "maximum number of sector\n"
               "retries were exceeded.  You may need to use the"
               " --unattended-rab option\n"
               "to increase the maximum limit.\n");
     else
        printf("\nUser aborted this disk copy.\n");
    }

 create_info_file();    
}

//==============================================================================
// Copy disk/image(s).
//
//   pass: void
// return: void
//==============================================================================
static void copy_disks (void)
{
 char user[100];
 char b_str[1000];
 char e_str[1000];
 char *p;
 int i;
 int l;
 int count;
 int res;

 overwrite_flag = -1;
 
 // set input and output types based on input and output names
 set_xtype_xfile(disk.ifile, disk.itype);
 set_xtype_xfile(disk.ofile, disk.otype); 

 // if no --detect is specified then set detect type to 'mbee' if floppy,
 // ntwdm or remote is the input driver
 if (! disk.detect && ((strcmp("remote", disk.itype) == 0) ||
#ifdef WIN32
 (strcmp("ntwdm", disk.itype) == 0)))
#else
 (strcmp("floppy", disk.itype) == 0)))
#endif
    format_detect_select("mbee");

 // check we have an input file name
 if (! disk.ifile[0])
    {
     printf(APPNAME": use '--if' option to specify an input file.\n");
     disk.error = 1;
    }

 // check we have an output file name.
 if (! disk.ofile[0])
    {
     printf(APPNAME": use '--of' option to specify an output file.\n");
     return;
    }

 // copy one disk then exit
 if (disk.count < 0)
    {
     strcpy(ofile_name, disk.ofile);
     if (strcmp(disk.ifile, ofile_name) == 0)
        {
         printf(APPNAME": input and output file names must be different!\n");
         return;
        }

     // check if destination file already exists and what to do
     if (! overwrite_permission(1, ofile_name))
        return;
     
     copy_one_disk();
     if (! disk.error && disk.verbose)
        {
         printf("\n");
         printf("Sector read errors: %d\n", sect_errors_tot);
         printf("       Retry total: %d\n", sect_retries_tot);
        } 
     close_files();
     return;
    }

 // prompt user for each disk to be copied (verbose has no affect here)
 count = disk.count;
 
 p = strstr(disk.ofile, ".");
 if (p)
    {
     strncpy(b_str, disk.ofile, (p - disk.ofile));
     b_str[(p - disk.ofile)] = 0;
     strcpy(e_str, p);
    }
 else
    {
     strcpy(b_str, disk.ofile);
     e_str[0] = 0;
    } 

 for (;;)
    {
     printf("\n");

     sprintf(ofile_name, "%s%03d%s", b_str, count, e_str);

     if (strcmp(disk.ifile, ofile_name) == 0)
        {
         printf(APPNAME": input and output file names must be different!\n");
         return;
        }

     l = (80 - (strlen(ofile_name) + 2)) / 2;
     for (i = 0; i < l; i++)
        putchar('=');
     printf(" %s ", ofile_name);
     for (i = 0; i < l; i++)
        putchar('=');
     printf("\n\n");

     // prompt the user to insert disk or exit
     user[0] = 0;
     while ((strcmp("\n", user) != 0) && (strcmp("e\n", user) != 0) &&
        (strcmp("E\n", user) != 0))
        {
         printf("Place next disk into drive '%s' and press ENTER or 'e'"
         " to exit ?: ", disk.ifile);
         if (fgets(user, sizeof(user)-2, stdin))
            {} // ignore result 
        }
     if (strcmp("\n", user) == 0)
        {
         // check if destination file already exists and what to do
         res = overwrite_permission(2, ofile_name);
         if (res == 1)
            {
             copy_one_disk();
             if (! disk.error)
                {
                 printf("\n");
                 printf("Sector read errors: %d\n", sect_errors_tot);
                 printf("       Retry total: %d\n", sect_retries_tot);
                }
             close_files();
            }
         else
            if (res == 2)
               return;
         count++;
        }
     else
        return;    
    } 
}

//==============================================================================
// Disk info.
//
// Dump geometry of the disk.
//
//   pass: void
// return: void
//==============================================================================
static void dump_geometry (void)
{
 // set input type based on input name
 set_xtype_xfile(disk.ifile, disk.itype); 

 // if no --detect is specified then set detect type to 'mbee' if floppy,
 // ntwdm or remote is the input driver
 if (! disk.detect && ((strcmp("remote", disk.itype) == 0) ||
#ifdef WIN32
 (strcmp("ntwdm", disk.itype) == 0)))
#else
 (strcmp("floppy", disk.itype) == 0)))
#endif 
    format_detect_select("mbee");

 if (open_drives() == -1)
    return;

 printf("\n");   

 if (! (xdg.dg_special & 0x01))
    printf("       Sidedness: %d (%s)\n"
           "       Cylinders: %d\n"
           "           Heads: %d\n"
           "   Sectors/track: %d\n"
           "     Sector base: %d\n"
           "     Sector size: %d\n"
           " Input data rate: %d (%s)\n"
           "Output data rate: %d (%s)\n"
           "         R/W gap: %d\n"
           "      Format gap: %d\n"
           "         FM mode: %d (%s)\n"
           "   No multi mode: %d (%s)\n"
           "     No skip del: %d (%s)\n"
           "       iside1as0: %d\n"
           "       secbase1s: %d\n"
           "       secbase2c: %d\n"
           "       secbase2s: %d\n"
           "       Read skew: %d\n"
           "Read skew offset: %d\n"
           "   Special flags: %d\n\n",
           dg.dg_sidedness, sidedness[dg.dg_sidedness],
           dg.dg_cylinders,
           dg.dg_heads,
           dg.dg_sectors,
           dg.dg_secbase,
           dg.dg_secsize,
           xdg.dg_idatarate, data_rates[xdg.dg_idatarate],
           xdg.dg_odatarate, data_rates[xdg.dg_odatarate],
           dg.dg_rwgap,
           dg.dg_fmtgap,
           dg.dg_fm, noyes[dg.dg_fm],
           dg.dg_nomulti, noyes[dg.dg_nomulti],
           dg.dg_noskip, noyes[dg.dg_noskip],
           xdg.dg_side1as0,
           xdg.dg_secbase1s,
           xdg.dg_secbase2c,
           xdg.dg_secbase2s,
           xdg.dg_skew_val,
           xdg.dg_skew_ofs,
           xdg.dg_special
          );
 else
    printf("       Sidedness: %d (%s)\n"
           "       Cylinders: %d (may vary)\n"
           "           Heads: %d\n"
           "   Sectors/track: 10x512 and 5x1024\n"
           "     Sector base: %d\n"
           "     Sector size: 512 and 1024\n"
           " Input data rate: %d (%s)\n"
           "Output data rate: %d (%s)\n"
           "         R/W gap: %d\n"
           "      Format gap: %d\n"
           "         FM mode: %d (%s)\n"
           "   No multi mode: %d (%s)\n"
           "     No skip del: %d (%s)\n"
           "       iside1as0: %d\n"
           "       secbase1s: %d\n"
           "       secbase2c: %d\n"
           "       secbase2s: %d\n"
           "       Read skew: %d\n"
           "Read skew offset: %d\n"
           "   Special flags: %d\n\n",
           dg.dg_sidedness, sidedness[dg.dg_sidedness],
           dg.dg_cylinders,
           dg.dg_heads,
           dg.dg_secbase,
           xdg.dg_idatarate, data_rates[xdg.dg_idatarate],
           xdg.dg_odatarate, data_rates[xdg.dg_odatarate],
           dg.dg_rwgap,
           dg.dg_fmtgap,
           dg.dg_fm, noyes[dg.dg_fm],
           dg.dg_nomulti, noyes[dg.dg_nomulti],
           dg.dg_noskip, noyes[dg.dg_noskip],
           xdg.dg_side1as0,
           xdg.dg_secbase1s,
           xdg.dg_secbase2c,
           xdg.dg_secbase2s,
           xdg.dg_skew_val,
           xdg.dg_skew_ofs,
           xdg.dg_special
          );

 close_files();
}

//==============================================================================
// Main
//
//   pass: int argc			number of arguments
//         char *argv[]			pointer to pointer to arguments
// return: int				0
//==============================================================================
int main (int argc, char *argv[])
{
 int res = EXIT_SUCCESS;
 
 // disable ctrl-C, ctrl-\ and ctrl-Z as it can leave floppy in unstable state
#ifdef WIN32
 // disables ctrl-C
 SetConsoleCtrlHandler(NULL, TRUE);
#else
 // disables ctrl-C, ctrl-\ and ctrl-Z
 signal(SIGINT,SIG_IGN);  // disable ctrl-C
 signal(SIGQUIT,SIG_IGN); // disable ctrl-backslash
 signal(SIGTSTP,SIG_IGN); // disable ctrl-Z
#endif 

 int exitstatus = 0;

 // initialise the functions module (before any console IO)
 functions_init();

 // check the operating system is the correct version
 if (check_os_version() == -1)
    return EXIT_FAILURE;
 
 // options initilisation
 options_init();

 // get the default home account directory.  This will typically be something
 // like c:\Program files\ubeedisk under windows and /home/username/.ubeedisk
 // under Unices if running from standard locations.
#ifdef WIN32
 int i;
 
 // get the path and executable filename
 if (GetModuleFileName(NULL, userhome, sizeof(userhome)) == 0)
    {
     printf(APPNAME": Unable to find the executable path\n");
     return -1;
    }

 // delete the executable name part and last '\' character
 i = strlen(userhome);
 while (userhome[--i] != SLASHCHAR)
    {}
 userhome[i] = 0;
 strcpy(userhome_path, userhome);
#else
 char *s;

 s = getenv("HOME");
 if (s == NULL)
    {
     printf(APPNAME": Unable to find the user's home path\n");
     return -1;
    }
 else
    {
     strcpy(userhome_path, s);
     snprintf(userhome, sizeof(userhome), "%s/."APPNAME, s);
    } 
#endif

 snprintf(userhome_confpath, sizeof(userhome_confpath), "%s%s",
 userhome, DIR_CONF);

 // process command line and initilisation options
 if (! exitstatus)
#ifdef WIN32 
    exitstatus = options_process(c_argc, c_argv);
#else    
    exitstatus = options_process(argc, argv);
#endif    

 if (exitstatus != 0)
    res = EXIT_FAILURE;
 else
    {
     dsk_reportfunc_set(report, report_end);

     switch (disk.disk)
        {
         case UBEEDISK_COPY : // copy disks
            copy_disks();
            break;
         case UBEEDISK_DUMP : // reserved for disk dump copy
            printf(APPNAME": disk dump option not implemented.\n");
            break;
         case UBEEDISK_GEOMETRY : // show detect format
            dump_geometry();
            break;
         case UBEEDISK_SCAN : // disk scan
            disk_scan();
            break;
         case UBEEDISK_FORMAT : // disk format
            disk_format();
            break;
         case UBEEDISK_SPEED : // disk speed test
            disk_speed();
            break;
         case UBEEDISK_CLEAN : // disk head clean
            disk_clean('c');
            break;
        }
    }

 // de-initialise the functions module
 functions_deinit();

 close_files();

 return res;
}
